# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import sys
import types
from textwrap import wrap
import re
import logging
import six
import importlib

# Define the logger
logger = logging.getLogger(__name__)

# Trait import
import traits.api

# Global parameters
_type_to_trait_id = {
    int: "Int",
    six.text_type: "Unicode",
    str: "Str",
    bytes: "Bytes",
    float: "Float"
}
# In order to convert nipype special traits, we define a dict of
# correspondances
_trait_cvt_table = {
    "InputMultiPath_TraitCompound": "List",
    "InputMultiPath": "List",
    "InputMultiObject": "List",
    "MultiPath": "List",
    "Dict_Str_Str": "DictStrStr",
    "OutputMultiPath_TraitCompound": "List",
    "OutputMultiPath": "List",
    "OutputMultiObject": "List",
    "OutputList": "List",
    "ImageFileSPM": "File",
}


def get_trait_desc(trait_name, trait, def_val=None, use_wrap=True):
    """ Generate a trait string description of the form:

    [parameter name: type (default trait value) string help (description)]

    Parameters
    ----------
    name: string (mandatory)
        the trait name
    trait: a trait instance (mandatory)
        a trait instance
    def_val: object (optional)
        the trait default value
        If not in ['', None] add the default trait value to the trait
        string description.
    use_wrap: bool (optional)
        if True, use text wrapping to 70 columns

    Returns
    -------
    manhelpstr: str
        the trait description.
    """
    # Get the trait description
    desc = trait.desc

    # Get the trait type
    trait_id = trait_ids(trait)

    if trait.output and isinstance(trait.trait_type,
                                   (traits.api.File, traits.api.Directory)):
        if trait.input_filename is False:
            trait_id[0] += ' (filename: output)'
        else:
            trait_id[0] += ' (filename: input)'

    # Add the trait name (bold)
    manhelpstr = ["{0}".format(trait_name)]

    # Get the default value string representation
    if def_val not in ["", None]:
        def_val = ", default value: {0}".format(repr(def_val))
    else:
        def_val = ""

    # Get the paramter type (optional or mandatory)
    if trait.optional:
        dtype = "optional"
    else:
        dtype = "mandatory"

    # Get the default parameter representation: trait type of default
    # value if specified
    line = "{0}".format(trait.info())
    #if not trait.output:
    line += " ({0} - {1}{2})".format(trait_id, dtype, def_val)

    # Wrap the string properly
    if use_wrap:
        manhelpstr = wrap(line, 70,
                          initial_indent=manhelpstr[0] + ": ",
                          subsequent_indent="    ")
    else:
        manhelpstr = [manhelpstr[0] + ": " + line]

    # Add the trait description if specified
    if desc:
        for line in desc.split("\n"):
            # line = re.sub("\s+", " ", line)
            if use_wrap:
                indent = ''
                s = line.strip()
                if s:
                    # keep text indentation
                    indent = line[:line.index(s[0])]
                wline = wrap(line, 70, initial_indent="    ",
                            subsequent_indent="    " + indent)
                if len(wline) == 0:
                    # don't skip empty lines
                    wline = ['']
            else:
                wline = ['    ' + line]
            manhelpstr += wline
    else:
        manhelpstr += wrap("No description.", 70,
                           initial_indent="    ",
                           subsequent_indent="    ")

    return manhelpstr


def is_trait_value_defined(value):
    """ Check if a trait value is valid.

    Parameters
    ----------
    value: type (mandatory)
        a value to test.

    Returns
    -------
    out: bool
        True if the value is valid,
        False otherwise.
    """
    # Initialize the default value
    is_valid = True

    # Check if the trait value is not valid
    if (value is None or value is traits.api.Undefined or
       (isinstance(value, six.string_types) and value == "")):

        is_valid = False

    return is_valid


def is_trait_pathname(trait):
    """ Check if the trait is a file or a directory.

    Parameters
    ----------
    trait: CTrait (mandatory)
        the trait instance we want to test.

    Returns
    -------
    out: bool
        True if trait is a file or a directory,
        False otherwise.
    """
    return (isinstance(trait.trait_type, traits.api.File) or
            isinstance(trait.trait_type, traits.api.Directory))


def trait_ids(trait, modules=set()):
    """Return the type of the trait: File, Enum etc...

    Parameters
    ----------
    trait: trait instance (mandatory)
        a trait instance
    modules: set (optional)
        modifiable set of modules names that should be imported to instantiate
        the trait

    Returns
    -------
    main_id: list
        the string description (type) of the input trait.
    """
    # Get the trait class name
    if hasattr(trait, 'handler'):
        handler = trait.handler or trait
    else:
        # trait is already a handler
        handler = trait
    main_id = handler.__class__.__name__
    if main_id == "TraitCoerceType":
        real_id = _type_to_trait_id.get(handler.aType)
        if real_id:
            main_id = real_id

    # Use the conversion table to normalize the trait id
    if main_id in _trait_cvt_table:
        main_id = _trait_cvt_table[main_id]

    # Debug message
    logger.debug("Trait with main id %s", main_id)

    # Search for inner traits
    inner_ids = []

    # Either case
    if main_id in ["Either", "TraitCompound"]:

        # Debug message
        logger.debug("A coumpound trait has been found %s", repr(
            handler.handlers))

        # Build each trait compound description
        trait_description = []
        for sub_trait in handler.handlers:
            if not isinstance(sub_trait, (traits.api.TraitType,
                                          traits.api.TraitInstance,
                                          traits.api.TraitCoerceType,
                                          traits.api.TraitHandler)):
                sub_trait = sub_trait()
            trait_description.extend(trait_ids(sub_trait, modules))
        return trait_description

    elif main_id == "Instance":
        inner_id = handler.klass.__name__
        mod = handler.klass.__module__
        if mod not in ("__builtin__", "__builtins__"):
            modules.add(mod)
            inner_id = '.'.join((mod, inner_id))
        return [main_id + "_" + inner_id]

    elif main_id == "TraitInstance":
        if handler.aClass is type(traits.api.Undefined):
            return ['Undefined']
        inner_id2 = _type_to_trait_id.get(handler.aClass)
        if inner_id2:
            return [inner_id2]
        inner_id = handler.aClass.__name__
        mod = handler.aClass.__module__
        if mod not in ("__builtin__", "__builtins__", "builtins"):
            modules.add(mod)
            inner_id = '.'.join((mod, inner_id))
        return [main_id + "_" + inner_id]

    elif handler is type(traits.api.Undefined):
        return ['Undefined']

    # Default case
    else:
        if not hasattr(handler, 'inner_traits'):
            return [main_id]

        # FIXME may recurse indefinitely if the trait is recursive
        inner_id = '_'.join((trait_ids(i, modules)[0]
                             for i in handler.inner_traits()))
        if not inner_id:
            klass = getattr(handler, 'klass', None)
            if klass is not None:
                inner_ids = [i.__name__ for i in klass.__mro__]
            else:
                inner_ids = []
        else:
            inner_ids = [inner_id]

        # Format the output string result
        if inner_ids:
            return [main_id + "_" + inner_desc for inner_desc in inner_ids]
        else:
            return [main_id]

def is_file_trait(trait, allow_dir=False, only_dirs=False):
    """
    Tells if the given trait is a File (and/or dict) or may be a file (for a
    compound trait)
    """
    ids = trait_ids(trait)
    if not only_dirs and any(['File' in x for x in ids]):
        return True
    if allow_dir and any(['Directory' in x for x in ids]):
        return True
    return False

def relax_exists_constraint(trait):
    """ Relax the exist constraint of a trait

    Parameters
    ----------
    trait: trait
        a trait that will be relaxed from the exist constraint
    """
    # If we have a single trait, just modify the 'exists' contrain
    # if specified
    if hasattr(trait.handler, "exists"):
        trait.handler.exists = False

    # If we have a selector, call the 'relax_exists_constraint' on each
    # selector inner components.
    main_id = trait.handler.__class__.__name__
    if main_id == "TraitCompound":
        for sub_trait in trait.handler.handlers:
            sub_c_trait = traits.api.CTrait(0)
            sub_c_trait.handler = sub_trait
            relax_exists_constraint(sub_c_trait)
    elif len(trait.inner_traits) > 0:
        for sub_c_trait in trait.inner_traits:
            relax_exists_constraint(sub_c_trait)



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import traits.api as traits
# Soma import
from soma.qt_gui.qt_backend import Qt
import sip


class NonEditableControlWidget(object):

    """ Displays that a control cannot be edited or seen. This is used for
    unknown types.
    """

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control value is correct.

        If the new entered value is not correct, the backroung control color
        will be red.

        Parameters
        ----------
        control_instance: QLineEdit (mandatory)
            the control widget we want to validate

        Returns
        -------
        out: bool
            True if the control value is valid,
            False otherwise
        """
        return True

    @classmethod
    def check(cls, control_instance):
        """ Check if a controller widget control is filled correctly.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        control_instance: QLineEdit (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, user_data=None):
        """ Method to create the string widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: str (mandatory)
            the default control value
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: QLineEdit,
            associated label: QLabel)
        """
        # Create the widget that will be used to fill a string
        try:
            str_value = str(control_value)
        except TypeError:
            try:
                str_value = repr(control_value)
            except TypeError:
                str_value = ' &lt;%s&gt;' % type(control_value).__name__

        widget = Qt.QLabel(
            '<style>background-color: gray; text-color: red;</style>'
            + str_value, parent)
        widget.setEnabled(False)

        # Create the label associated with the string widget
        control_label = getattr(trait, 'label', control_name)
        if label_class is None:
            label_class = Qt.QLabel
        if control_label is not None:
            label = label_class(control_label, parent)
        else:
            label = None

        return (widget, label)

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          reset_invalid_value=False, *args, **kwargs):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        pass

    @staticmethod
    def update_controller_widget(controller_widget, control_name,
                                 control_instance):
        """ Update one element of the controller widget.

        At the end the controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QWidget (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        try:
            test = control_instance.setText
        except ReferenceError:
            # widget deleted in the meantime
            return

        if sip.isdeleted(control_instance.__init__.__self__):
            NonEditableControlWidget.disconnect(
                controller_widget, control_name, control_instance)
            return

        # Get the trait value
        new_controller_value = getattr(
            controller_widget.controller, control_name, traits.Undefined)
        try:
            str_value = str(new_controller_value)
        except TypeError:
            try:
                str_value = repr(new_controller_value)
            except TypeError:
                str_value = ' &lt;%s&gt;' % type(new_controller_value).__name__
        control_instance.setText(str_value)

    @classmethod
    def connect(cls, controller_widget, control_name, control_instance):
        """ Connect a 'Str' or 'String' controller trait and a
        'StrControlWidget' controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        pass


    @staticmethod
    def disconnect(controller_widget, control_name, control_instance):
        """ Disconnect a 'Str' or 'String' controller trait and a
        'StrControlWidget' controller widget control.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        pass



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import logging
from functools import partial
import traits.api as traits
import sys
import six

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.utils.functiontools import SomaPartial
from soma.qt_gui.timered_widgets import TimeredQLineEdit
from soma.utils.weak_proxy import weak_proxy
from .Str import StrControlWidget


class BytesControlWidget(StrControlWidget):

    """ Control to enter a bytes string.
    """

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control value is correct.

        If the new entered value is not correct, the backroung control color
        will be red.

        Parameters
        ----------
        control_instance: QLineEdit (mandatory)
            the control widget we want to validate

        Returns
        -------
        out: bool
            True if the control value is valid,
            False otherwise
        """
        # Get the current control palette
        control_palette = control_instance.palette()

        # Get the control current value
        control_value = control_instance.value()

        color = QtCore.Qt.white
        red = QtGui.QColor(255, 220, 220)
        yellow = QtGui.QColor(255, 255, 200)

        # If the control value is not empty, the control is valid and the
        # backgound color of the control is white
        is_valid = False

        if control_value in (b'', None, traits.Undefined):
            if control_instance.optional:
                # If the control value is optional, the control is valid and
                # the backgound color of the control is yellow
                color = yellow
                is_valid = True
            else:
                color = red
                if control_value != b'':
                    # allow to reset value
                    is_valid = True

        else:
            is_valid = True

        # Set the new palette to the control instance
        control_palette.setColor(control_instance.backgroundRole(), color)
        control_instance.setPalette(control_palette)

        return is_valid

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          reset_invalid_value=False, *args, **kwargs):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QLineEdit (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Update the controller only if the control is valid
        if BytesControlWidget.is_valid(control_instance):

            # Get the control value
            new_trait_value = control_instance.value()
            if new_trait_value not in (traits.Undefined, None):
                new_trait_value = six.ensure_binary(new_trait_value)

            # value is manually modified: protect it
            if getattr(controller_widget.controller, control_name) \
                    != new_trait_value:
                controller_widget.controller.protect_parameter(control_name)
            # Set the control value to the controller associated trait
            setattr(controller_widget.controller, control_name,
                    new_trait_value)
            logger.debug(
                "'BytesControlWidget' associated controller trait '{0}' has "
                "been updated with value '{1}'.".format(
                    control_name, new_trait_value))
        elif reset_invalid_value:
            # invalid, reset GUI to older value
            old_trait_value = getattr(controller_widget.controller,
                                      control_name)
            control_instance.set_value(old_trait_value)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function

import six
from importlib import import_module
from pkgutil import iter_modules


def find_subclasses_in_module(module_name, parent_class):
    '''
    Finds all the classes defined in a Python module that derive from a
    given class or class name. If the module is a package, it also look
    into submodules.
    '''
    if isinstance(parent_class, six.string_types):
        check = lambda item: (isinstance(item, type) and 
                              item.__module__ == module_name and 
                              parent_class in (i.__name__ 
                                               for i in item.__mro__))
    else:
        check = lambda item: (isinstance(item, type) and 
                              item.__module__ == module_name and 
                              issubclass(item, parent_class))
    for i in find_items_in_module(module_name, check):
        yield i


def find_items_in_module(module_name, check):
    '''
    Finds all the items defined in a Python module that where *check(cls)* is
    *True*. If the module is a package, it also look recursively into
    submodules.
    '''
    try:
        module = import_module(module_name)
    except ImportError:
        return
    
    for i in six.itervalues(module.__dict__):
        if check(i):
            yield i
    path = getattr(module, '__path__', None)
    if path:
        for importer, submodule_name, ispkg in iter_modules(path):
            for j in find_items_in_module('%s.%s' % 
                    (module.__name__, submodule_name), check):
                yield j


class ClassFactory(object):
    '''
    *ClassFactory* is the base class for creating factories that can look
    for classes in Python modules and create instances.
    '''
    
    def __init__(self, class_types={}):
        # List of Python modules where classes are looked for
        self.module_path = []
        # Instance-level association between a class_type (which is a string)
        # and the corresponding class. There can also be class-level
        # class_types (see get_class method).
        self.class_types = {}
        # Cache of instances returned by get method. This is a dictionary
        # whose keys are (class_type, factory_id) and values are instances.
        self.instances = {}
    
    def find_class(self, class_type, factory_id):
        '''
        Finds a class deriving of the class corresponding to *class_type* and
        whose *factory_id* attribute has a given value. Look for all
        subclasses of parent class declared in the modules of
        :attr:`self.module_path`
        and returns the first one having ``cls.factory_id == factory_id``. If
        none is found, returns *None*.
        '''
        base_class = self.get_class(class_type)
        for module_name in self.module_path:
            for cls in find_subclasses_in_module(module_name, base_class):
                if cls.factory_id == factory_id:
                    return cls
        return None
    
    def get_class(self, class_type):
        '''
        Returns the class corresponding to a given class type. First look
        for a dictionary in self.class_types, then look into parent
        classes.
        '''
        class_types = getattr(self, 'class_types', None)
        if class_types:
            cls = class_types.get(class_type)
            if cls is not None:
                return cls
        for parent_class in self.__class__.__mro__:
            class_types = getattr(parent_class, 'class_types', None)
            if class_types:
                cls = class_types.get(class_type)
                if cls is not None:
                    return cls
        raise ValueError('Unknown class type: %s' % class_type)



    def get(self, class_type, factory_id):
        '''
        Returns an instance of the class identified by *class_type* and
        *factory_id*. There can be only one instance per *class_type* and
        *factory_id*. Once created with the first call of this method, it
        is stored in :attr:`self.instances` and simply returned in subsequent
        calls.

        If *get_class* is *True*, returns the class instead of an instance
        '''
        instance = self.instances.get((class_type, factory_id))
        if instance is None:
            cls = self.find_class(class_type, factory_id)
            if cls is not None:
                if hasattr(cls.__init__, '__code__') \
                        and cls.__init__.__code__.co_argcount != 1:
                    # The class constructor takes arguments: we cannot
                    # instantiate it: register the class itself
                    instance = cls
                else:
                    instance = cls()
                self.instances[(class_type, factory_id)] = instance
            else:
                raise ValueError('Cannot find a class for class type "%s" '
                                 'and factory id "%s"' % (class_type, 
                                                          factory_id))
        return instance



# -*- coding: utf-8 -*-
# Copyright CEA and IFR 49 (2000-2005)
#
#  This software and supporting documentation were developed by
#      CEA/DSV/SHFJ and IFR 49
#      4 place du General Leclerc
#      91401 Orsay cedex
#      France
#
# This software is governed by the CeCILL license version 2 under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL license version 2 as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license version 2 and that you accept its terms.

from __future__ import print_function
from __future__ import absolute_import
from soma.qt_gui.qt_backend.QtGui import *
from soma.qt_gui.qt_backend.QtCore import QSize
import re
import os
import six
from six.moves import range


class GenericTableEditor(QWidget):
    ''' Table viewer / editor widget
    '''

    def __init__(self, file=None, context=None, parent=None):
        QWidget.__init__(self, parent)
        self.layout = QVBoxLayout(self)
        self.context = context
        self.params = None
        self.cell2tab = None
        # - ------------------------------------------------------------- - #
        self.working_dir = '.'
        self.precision = 4
        self.buf_table = None
        #
        # - the menu bar
        self.menuBar = QMenuBar(self)
        self.layout.addWidget(self.menuBar)
        self.menuBar.show()

        # - File menu
        self.menu_file = self.menuBar.addMenu("&File")

        # - - open
        self.menu_file.addAction("&Open", self.menu_open_action)

        # - - save_as
        self.menu_file.addAction("&Save as", self.menu_save_as_action)

        # - Edit menu
        self.menu_edit = self.menuBar.addMenu("&Edit")

        # - - select
        self.menu_file.addAction("&Select", self.menu_select_action)
        # - - replace
        self.menu_file.addAction("&Replace", self.menu_replace_action)
        # - - add col
        self.menu_file.addAction("&Add col", self.menu_addCol)
        # - - remove col
        self.menu_file.addAction("&Remove &col(s)", self.menu_removeCols)
        # - - add line
        self.menu_file.addAction("A&dd raw(s)", self.menu_addLine)
        # - - remove line
        self.menu_file.addAction("Remove &raw(s)", self.menu_removeLines)

        # - Settings menu
        self.menu_settings = self.menuBar.addMenu("&Settings")
        # - -
        self.menu_settings.addAction("&Parameters", self.showParamsBox)

        # - - View menu
        self.menu_view = self.menu_settings.addMenu("&View")
        self.precisionButtonGrp = QActionGroup(self)
        self.precisionButtonGrp.setExclusive(True)

        self.precision_full = QAction('Full precision', self.menu_view)
        self.precision_full.setObjectName('Precision1')
        self.precision_full.setCheckable(True)

        self.precision_4 = QAction('10^-4', self.menu_view)
        self.precision_4.setObjectName('Precision2')
        self.precision_4.setCheckable(True)
        self.precision_4.setChecked(True)

        self.precision_2 = QAction('10^-2', self.menu_view)
        self.precision_2.setObjectName('Precision')
        self.precision_2.setCheckable(True)

        self.precisionButtonGrp.addAction(self.precision_full)
        self.precisionButtonGrp.addAction(self.precision_2)
        self.precisionButtonGrp.addAction(self.precision_4)
       # self.precision_full.show()

        self.menu_view.addAction(self.precision_full)
        self.menu_view.addAction(self.precision_4)
        self.menu_view.addAction(self.precision_2)

        self.precisionButtonGrp.triggered.connect(self.setPrecision)

        # - Tool menu
        # self.menu_tool=self.menuBar.addMenu("&Tools")

        # -
        # self.menu_tool.addAction("&Convert cell to table",
        # self.showCell2TabBox)

        # - ------------------------------------------------------------- - #
        # - Table
        self.gui_table = QTableWidget(self)
        self.gui_table.setSortingEnabled(False)
        self.layout.addWidget(self.gui_table)
        self.gui_table.cellChanged.connect(self.guiVal2BufVal)
        self.gui_table.horizontalHeader().sectionClicked.connect(
            self.buf_table_sort)

        # - ------------------------------------------------------------- - #
        # - Operation
        self.operations_widget = QWidget(self)
        self.operations_layout = QVBoxLayout(self.operations_widget)
        self.operations_widget.setLayout(self.operations_layout)
        self.operations_layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.operations_widget)

        ope_sort = QPushButton("Sort", self.operations_widget)
        self.operations_layout.addWidget(ope_sort)
        ope_sort.clicked.connect(self.buf_table_sort)
        if file:
            self.load_from_file(file)

    def buf_table_sort(self, col=None, ascending=True):
        self.current_col_index = 0
        self.current_col_index = self.getSelectedCols()[0]
        self.buf_table.sort(self.order)
        self.gui_table.setSortingEnabled(False)
        self.buf_2_gui()

    def clear(self):
        self.buf_table = None
        self.data = None
        self.fileStructure = None
        self.col_names = None
        self.col_types = None
        self.setWindowTitle('No File loaded')
        self.buf_2_gui()

    def load_from_file(self, file=None, fileStructure=None):
        ''' This function needs the pandas module
        '''
        if file:
            self.file = file
        self.fileStructure = fileStructure

        # - clear the table
        self.gui_table.clear()
        import pandas as pd
        print(self.fileStructure)
        self.data = pd.read_csv(self.file)
        if self.data.shape[1] <= 1:
            self.data = pd.read_csv(self.file, sep=";")
        if self.data.shape[1] <= 1:
            self.data = pd.read_csv(self.file, sep="\t")
        if self.data.shape[1] <= 1:
            self.data = pd.read_csv(self.file, sep=" ")
        self.col_names = self.data.columns.tolist()

        col_types = pd.Series([self.data[col].dtype for col in self.data])
        col_types = col_types.replace(
            {"object": "varchar", "float64": "float", "int64": "integer"})
        self.col_types = [t for t in col_types]
        self.buf_table = [l for l in self.data.itertuples(index=False)]
        self.gui_table.setColumnCount(self.data.shape[1])
        self.gui_table.setRowCount(self.data.shape[0])

        # GenericTableEditor.convTab2NativType(self.buf_table, self.col_types)
        self.buf_2_gui()

        # set the main window name
        self.setWindowTitle(os.path.basename(self.file))
        if self.parent() is not None:
            self.parent().setWindowTitle(os.path.basename(self.file))

    def numCols(self):
        if self.col_types:
            return len(self.col_types)
        else:
            return 0

    def numRows(self):
        if self.buf_table:
            return len(self.buf_table)
        else:
            return 0

    def buf_2_gui(self):
        # disable auto sorting temporarily because we need to freeze
        # the items position while inserting them
        sortstate = self.gui_table.isSortingEnabled()
        self.gui_table.setSortingEnabled(False)
        # udpate gui table size if needed
        if self.gui_table.columnCount() != self.numCols():
            self.gui_table.setColumnCount(self.numCols())
        if self.gui_table.rowCount() != self.numRows():
            self.gui_table.setRowCount(self.numRows())

        # update gui header
        if self.col_names:
            self.gui_table.setHorizontalHeaderLabels(self.col_names)

        # copy buf data into gui
        if self.buf_table:
            self.gui_table.blockSignals(True)
            for i in range(len(self.buf_table)):
                for j in range(len(self.buf_table[i])):
                    val = self.buf_table[i][j]
                    if self.precision and isinstance(val, float):
                        val = round(val, self.precision)
                    elif val is None:
                        val = '-'
                    it = QTableWidgetItem(str(val))
                    self.gui_table.setItem(i, j, it)
            self.gui_table.blockSignals(False)
        self.gui_table.setSortingEnabled(sortstate)

    def order(self, a, b):
        va = a[self.current_col_index]
        vb = b[self.current_col_index]
        if va < vb:
            return -1
        if va > vb:
            return +1
        return 0

    def menu_open_action(self):
        file = str(QFileDialog.getOpenFileName(
                   self, "", "", "", None, QtGui.QFileDialog.DontUseNativeDialog))
        if file and file != '':
            self.load_from_file(file=file)

    def menu_save_as_action(self):
        file = str(QFileDialog.getSaveFileName(
            self, "", "", "", None, QtGui.QFileDialog.DontUseNativeDialog))
        self.data.to_csv(file, index=False)

    def menu_select_action(self):
        selection = []
        for item in self.gui_table.selectedItems():
            selection.append(str(item.text()))
        # send an 'inputSelection' event with the selection as paramater
        if self.context:
            self.context.event('inputSelection', selection)

    def menu_replace_action(self):
        try:
            from_str = str(
                QInputDialog.getText(
                    self, 'Enter regular expression',
                    'Enter query to replace', QLineEdit.Normal, '')[0])
            if not from_str:
                return None
            from_re = re.compile(from_str)
            to_str = str(QInputDialog.getText(
                self, 'Enter regular expression',
                'Replace ' + from_str + ' with', QLineEdit.Normal, '')[0])
            if not to_str:
                return None
            for tSelec in self.gui_table.selectedRanges():
                for row in six.moves.xrange(tSelec.topRow(), tSelec.bottomRow() + 1):
                    for col in six.moves.xrange(tSelec.leftColumn(), tSelec.rightColumn() + 1):
                        # regexp.sub(  	replacement, string
                        val = str(self.buf_table[row][col])
                        # if type(val)==types.StringType : val = str(val)
                        newval = from_re.sub(to_str, val)
                        self.setBufValFromStr(row, col, newval)
            self.buf_2_gui()
        except Exception as e:
            QMessageBox.warning(self, "WARNING", str(e))
            return None

    def menu_addCol(self):
        name = str(QInputDialog.getText(self, 'Add column',
                                        'Enter column name', QLineEdit.Normal, '')[0])
        if not name:
            return None
        # add an item for each line in the buf tab
        for line in self.buf_table:
            line.append(0)
        if self.col_names:
            self.col_names.append(name)
        self.buf_2_gui()

    def menu_removeCols(self):
        selectedCols = self.getSelectedCols()
        selectedCols.sort()
        # remove in table
        nbDeleted = 0
        for c in selectedCols:
            for line in self.buf_table:
                del(line[c - nbDeleted])
            if self.col_names:
                del(self.col_names[c - nbDeleted])
            del(self.col_types[c - nbDeleted])
            nbDeleted += 1
        self.buf_2_gui()

    def menu_addLine(self):
        nbline = QInputDialog.getInteger(self, 'Add line(s)',
                                         'Enter number line(s) to add',  1, 1, 2147483647, 1)[0]
        for i in six.moves.xrange(nbline):
            self.buf_table.append([0] * len(self.numCols()))
        self.buf_2_gui()

    def menu_removeLines(self):
        selectedRows = self.getSelectedRows()
        selectedRows.sort()
        # remove in table
        nbDeleted = 0
        for r in selectedRows:
            del(self.buf_table[r - nbDeleted])
            nbDeleted += 1
        self.buf_2_gui()

    def getSelectedCols(self):
        ret = []
        for tSelec in self.gui_table.selectedRanges():
            for col in six.moves.xrange(tSelec.leftColumn(), tSelec.rightColumn() + 1):
                ret.append(col)
        return ret

    def getSelectedRows(self):
        ret = []
        for tSelec in self.gui_table.selectedRanges():
            for raw in six.moves.xrange(tSelec.topRow(), tSelec.bottomRow() + 1):
                ret.append(raw)
        return ret

    def getSelectedCells(self):
        selections = []
        # print
        # 'numSelections:',self.genericTableEditor.gui_table.numSelections ()
        for selectnb in self.gui_table.selectedRanges():
            selections.append(self.gui_table.selection(selectnb))

        selectedCells = []
        selectedCellsContent = []
        for s in self.gui_table.selectedRanges():
            for row in six.moves.xrange(s.topRow(), s.bottomRow() + 1):
                for col in six.moves.xrange(s.leftColumn(), s.rightColumn() + 1):
                    cell = [row, col]
                    if not cell in selectedCells:
                        selectedCells.append(cell)
                        selectedCellsContent.append(
                            str(self.gui_table.item(row, col).text()))
        return selectedCellsContent

    def setBufValFromStr(self, row, col, valstr):
        """ Set a value and the apropiate type in the data table from a string"""
        typeConversionRequired = False
        if self.col_types[col] == 'integer':
            try:
                self.buf_table[row][col] = int(valstr)
            except ValueError:
                try:
                    typeConversionRequired = True
                    self.buf_table[row][col] = float(valstr)
                    self.col_types[col] = 'float'
                except ValueError:
                    self.buf_table[row][col] = valstr
                    self.col_types[col] = 'varchar'
        elif self.col_types[col] == 'float':
            try:
                self.buf_table[row][col] = float(valstr)
            except ValueError:
                typeConversionRequired = True
                self.buf_table[row][col] = valstr
                self.col_types[col] = 'varchar'
        else:
            self.buf_table[row][col] = valstr
        if typeConversionRequired:
            GenericTableEditor.convTab2NativType(self.buf_table,
                                                 self.col_types)

    def guiVal2BufVal(self, row, col):
        guiVal = str(self.gui_table.item(row, col).text())
        self.setBufValFromStr(row, col, guiVal)

    def getMenuBar(self): return self.menuBar

    def showParamsBox(self):
        if not self.params:
            self.params = GenericTableEditor.Parameters(self)
        self.params.initFields(self.fileStructure)
        self.params.show()

    def setPrecision(self, button):
        if button is self.precision_full:
            self.precision = None
        elif button is self.precision_4:
            self.precision = 4
        elif button is self.precision_2:
            self.precision = 2
        else:
            print('unknown precision')
        self.buf_2_gui()

    @staticmethod
    def convTab2NativType(data, colTypes):
        for col in range(len(colTypes)):
            if colTypes[col] == 'integer':
                for tuple in data:
                    if tuple[col]:
                        tuple[col] = int(tuple[col])
            elif colTypes[col] == 'float':
                for tuple in data:
                    if tuple[col]:
                        tuple[col] = float(tuple[col])
            elif colTypes[col] == 'varchar':
                for tuple in data:
                    if tuple[col]:
                        tuple[col] = str(tuple[col])


    class Parameters(QMainWindow):

        def __init__(self, tableEditor):
            QMainWindow.__init__(self, tableEditor)
            self.setWindowTitle("File reader/writer parameters")
            self.genericTableEditor = tableEditor

        def initFields(self, fileStructure):
            self.fileStructure = fileStructure
            # print('Parameters',self.fileStructure)

            self.widget = QWidget(self)
            widgetLayout = QVBoxLayout()
            self.widget.setLayout(widgetLayout)
            self.setCentralWidget(self.widget)
            # if not fileStructure : fileStructure={}

            # - Header
            self.header = QCheckBox('Header', self.widget)
            widgetLayout.addWidget(self.header)
            if self.fileStructure is not None \
                    and 'header' in self.fileStructure:
                header = self.fileStructure['header']
                if header:
                    self.header.setChecked(True)
                else:
                    self.header.setChecked(False)
                    # self.header.show()

            # - Quoted
            self.quoted = QCheckBox('Quoted', self.widget)
            widgetLayout.addWidget(self.quoted)
            if self.fileStructure is not None \
                    and 'quoted' in self.fileStructure:
                self.quoted.setChecked(self.fileStructure['quoted'])

            # - Speparators
            separator = QWidget(self.widget)
            widgetLayout.addWidget(separator)
            separatorLayout = QHBoxLayout()
            separator.setLayout(separatorLayout)
            separatorLayout.addWidget(QLabel('Separator'))
            self.separator = QComboBox()
            separatorLayout.addWidget(self.separator)
            self.separator.setObjectName('Separator')
            self.separator.addItem('""')
            self.separator.addItem('" "')
            self.separator.addItem('"\t"')
            self.separator.addItem('";"')
            self.separator.addItem('","')
            self.separator.setEditable(True)

            if self.fileStructure is not None \
                    and 'separator' in self.fileStructure:
                sep = self.fileStructure['separator']
                if not sep:
                    sep = ''
                self.separator.setEditText('"' + sep + '"')
            # self.displayColsNamesTypes()

            # - Updates
            update = QPushButton("Update", self.widget)
            widgetLayout.addWidget(update)
            update.clicked.connect(self.clickupdate)

        def displayColsNamesTypes(self):
            # - Header names and types
            columns = QWidget(self.widget)
            columnsLayout = QHBoxLayout()
            colums.setLayout(columnsLayout)
            self.columns = []
            labels = QWidget(columns)
            columnsLayout.addWidget(labels)
            labelsLayout = QVBoxLayout()
            labels.setLayout(labelsLayout)
            labelsLayout.addWidget(QLabel('Header'))
            labelsLayout.addWidget(QLabel('Types'))

            # - Get the number of columns from the gui
            try:
                nbCols = int(str(self.nbCols.text()))
            except ValueError:
                nbCols = 0

            if self.fileStructure is not None \
                    and 'header' in self.fileStructure:
                header = self.fileStructure['header']
            else:
                header = False

            if self.fileStructure is not None \
                    and 'types' in self.fileStructure:
                types = self.fileStructure['types']
            else:
                types = None

            class TypesQComboBox(QComboBox):

                def __init__(self, parent, type):
                    QComboBox.__init__(self, parent)
                    self.addItem('integer')
                    self.addItem('float')
                    self.addItem('varchar')
                    if type:
                        self.setEditText(type)

            for col in six.moves.xrange(nbCols):  # self.fileStructure['nbCols']):
                current = QWidget(columns)
                columnsLayout.addWidget(current)
                currentLayout = QVBoxLayout()
                current.setLayout(currentLayout)
                if header:
                    try:
                        lab = header[col]
                    except IndexError:
                        lab = ''
                else:
                    lab = ''
                if types:
                    try:
                        type = types[col]
                    except IndexError:
                        types = None
                else:
                    types = None

                self.columns.append(
                    (QLineEdit(lab, current), TypesQComboBox(current, type)))

        def clickupdate(self):
            # Rebuild the structFile from GUI
            structFile = {}
            # - Header (If header checked, take the header tab)
            header = False
            if self.header.isChecked():
                header = True

            # - Quoted
            quoted = self.quoted.isChecked()

            # - Separator
            separator = str(self.separator.currentText()).replace('"', '')

            structFile = {'header': header, 'quoted': quoted,
                          'separator': separator}  # , 'nbCols': nbCols}
            self.genericTableEditor.load_from_file(
                file=None, fileStructure=structFile)



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import logging
import six
import sys

# Define the logger
logger = logging.getLogger(__name__)

# Trait import
from traits.api import HasTraits, Event, CTrait, Instance, Undefined, \
    TraitType, TraitError, Any, Set, TraitInstance, TraitCoerceType, Tuple, \
    TraitHandler
import traits.api as traits
# Soma import
from soma.sorted_dictionary import SortedDictionary, OrderedDict
from soma.controller.trait_utils import _type_to_trait_id


class Controller(HasTraits):

    """ A Controller contains some traits: attributes typing and observer
    (callback) pattern.

    The class provides some methods to add/remove/inspect user defined traits.

    Attributes
    ----------
    `user_traits_changed` : Event
        single event that can be sent when several traits changes. This event
        has to be triggered explicitely to take into account changes due to
        call(s) to add_trait or remove_trait.

    Methods
    -------
    user_traits
    is_user_trait
    add_trait
    remove_trait
    _clone_trait
    """

    # This event is necessary because there is no event when a trait is
    # removed with remove_trait and because it is sometimes better to send
    # a single event when several traits changes are done (especially
    # when GUI is updated on real time). This event have to be triggered
    # explicitely to take into account changes due to call(s) to
    # add_trait or remove_trait.
    user_traits_changed = Event

    def __init__(self, *args, **kwargs):
        """ Initilaize the Controller class.

        During the class initialization create a class attribute
        '_user_traits' that contains all the class traits and instance traits
        defined by user (i.e.  the traits that are not automatically
        defined by HasTraits or Controller). We can access this class
        parameter with the 'user_traits' method.

        If user trait parameters are defined directly on derived class, this
        procedure call the 'add_trait' method in order to not share
        user traits between instances.
        """
        # Inheritance
        super(Controller, self).__init__(*args, **kwargs)

        # Create a sorted dictionnary with user parameters
        # The dictionary order correspond to the definition order
        self._user_traits = SortedDictionary()

        # Get all the class traits
        class_traits = self.class_traits()

        # If some traits are defined on the controller, create a list
        # with definition ordered trait name. These names will correspond
        # to user trait sorted dictionary keys
        if class_traits:
            
            sorted_names = []
            for name, trait in six.iteritems(class_traits):
                if self.is_user_trait(trait):
                    if getattr(trait, 'order', None):
                        # Only if trait.order exists AND trait.order is no None
                        sorted_names.append((getattr(trait, 'order'), name))
                    else:
                        sorted_names.append((-1, name))
                    
            sorted_names = [sorted_name[1] for sorted_name in sorted(sorted_names)]

            # Go through all trait names that have been ordered
            for name in sorted_names:

                # If the trait is defined on the class, need to clone
                # the class trait and add the cloned trait to the instance.
                # This step avoids us to share trait objects between
                # instances.
                if name in self.__base_traits__:
                    logger.debug("Add class parameter '{0}'.".format(name))
                    trait = class_traits[name]
                    self.add_trait(name, self._clone_trait(trait))

                # If the trait is defined on the instance, just
                # add the user parameter to the '_user_traits' instance
                # parameter
                else:
                    logger.debug("Add instance parameter '{0}'.".format(name))
                    self._user_traits[name] = class_traits[name]

    #
    # Private methods
    #

    def _clone_trait(self, clone, metadata=None):
        """ Creates a clone of a specific trait (ie. the same trait
        type but different ids).

        Parameters
        ----------
        clone: CTrait (mandatory)
            the input trait to clone.
        metadata: dict (opional, default None)
            some metadata than can be added to the trait __dict__.

        Returns
        -------
        trait: CTrait
            the cloned input trait.
        """
        # Create an empty trait
        trait = CTrait(0)

        # we need a CTrait, not a TraitType
        if isinstance(clone, TraitType):
            clone = clone.as_ctrait()

        # Clone the input trait in the empty trait structure
        trait.clone(clone)

        # Set the input trait __dict__ elements to the cloned trait
        # __dict__
        if clone.__dict__ is not None:
            trait.__dict__ = clone.__dict__.copy()

        # Update the cloned trait __dict__ if necessary
        if metadata is not None:
            trait.__dict__.update(metadata)

        return trait

    def _propagate_optional_parameter(self, trait, optional=None):
        """
        """
        # Get the trait class name
        if hasattr(trait, 'handler'):
            handler = trait.handler or trait
        else:
            handler = trait # hope it is already a handler
        main_id = handler.__class__.__name__
        if main_id == "TraitCoerceType":
            real_id = _type_to_trait_id.get(handler.aType)
            if real_id:
                main_id = real_id

        # Debug message
        logger.debug("Propagation optional parameter of trait with main id %s",
                     main_id)

        # Get the optional parameter and set the default value if necessary
        if optional is not None:
            trait.optional = optional
        else:
            optional = trait.optional
            if optional is None:
                optional = False
                trait.optional = optional

        # Either case
        if main_id in ["Either", "TraitCompound"]:

            # Debug message
            logger.debug("A coumpound trait has been found %s", repr(
                handler.handlers))

            # Update each trait compound optional parameter
            for sub_trait in handler.handlers:
                if not isinstance(sub_trait, (TraitInstance, TraitCoerceType,
                                              TraitHandler)):
                    sub_trait = sub_trait()
                self._propagate_optional_parameter(sub_trait, optional)

        # Default case
        else:
            # FIXME may recurse indefinitely if the trait is recursive
            for inner_trait in handler.inner_traits():
                self._propagate_optional_parameter(inner_trait, optional)

    #
    # Public methods
    #

    def user_traits(self):
        """ Method to access the user parameters.

        Returns
        -------
        out: dict
            a dictionnary containing class traits and instance traits
            defined by user (i.e.  the traits that are not automatically
            defined by HasTraits or Controller). Returned values are
            sorted according to the 'order' trait meta-attribute.
        """
        return self._user_traits

    def is_user_trait(self, trait):
        """ Method that evaluate if a trait is a user parameter
        (i.e. not an Event).

        Returns
        -------
        out: bool
            True if the trait is a user trait,
            False otherwise.
        """
        return not isinstance(trait.handler, Event)


    @staticmethod
    def checked_trait(trait):
        """ Check the trait and build a new one if needed.

        This function mainly checks the default value of the given trait,
        and tests in some ways whether it is valid ot not. If not, a new
        trait is created to replace it.

        For now it just checks that lists with a non-null minlen will actually
        get a default value which is a list with this minimum size. Otherwise
        it causes exceptions in the traits notification system at some point.

        Parameters
        ----------
        trait: Trait instance to be checked

        Returns
        -------
        new_trait: Trait instance
            the returned trait may be the input one (trait), or a new one if
            it had to be modified.
        """
        ut = getattr(trait, 'trait_type', trait)
        if isinstance(ut, traits.List):
            if ut.minlen != 0 and (not isinstance(ut.default, list)
                                   or len(ut.default) < ut.minlen):
                # default value is not OK, we have to build another one
                if isinstance(ut.default, list):
                    default = list(ut.default)
                else:
                    default = []
                default += [ut.item_trait.default] * (ut.minlen - len(default))
                old_trait = trait
                trait = traits.List(ut.item_trait, default, minlen = ut.minlen,
                                    maxlen=ut.maxlen)
                # copy metadata
                builtin = {'copy', 'type'}
                for meta, value in old_trait.__dict__.items():
                    if meta not in builtin:
                        setattr(trait, meta, value)
        return trait


    def add_trait(self, name, *trait):
        """ Add a new trait.

        Parameters
        ----------
        name: str (mandatory)
            the trait name.
        trait: traits.api (mandatory)
            a valid trait.
        """
        # Debug message
        logger.debug("Adding trait '{0}'...".format(name))

        # check trait default value inconsistencies
        trait = (self.checked_trait(trait[0]), ) + trait[1:]

        # Inheritance: create the instance trait attribute
        super(Controller, self).add_trait(name, *trait)

        # Get the trait instance and if it is a user trait load the traits
        # to get it direcly from the instance (as a property) and add it
        # to the class '_user_traits' attributes
        trait_instance = self.trait(name)
        if self.is_user_trait(trait_instance):
            #trait_instance.defaultvalue = trait_instance.default
            #try:
                #self.get(name)
            #except TraitError:
                ## default value is invalid
                #try:
                    #setattr(self, name, Undefined)
                #except TraitError:
                    ## Undefined is invalid, too...
                    #pass
            self._user_traits[name] = trait_instance

        # Update/set the optional trait parameter
        self._propagate_optional_parameter(trait_instance)

        # validate default value, or try to set another one
        new_trait = self.trait(name)
        if not isinstance(new_trait.trait_type, traits.Event):
            try:
                values = (getattr(self, name), traits.Undefined, None, '', 0)
            except TraitError:
                values = (traits.Undefined, None, '', 0)
            for value in values:
                try:
                    # validate() doesn't accept Undefined values when the
                    # "real" trait does. so we must really setattr()
                    #new_trait.validate(self, name, value)
                    setattr(self, name, value)
                    break  # OK
                except (traits.TraitError, TypeError) as e:
                    pass
            #else:
                ## should it be silent ?
                #print('value %s is invalid for %s.%s'
                      #% (repr(values[0]), repr(self), name), file=sys.stderr)

        self.user_traits_changed = True

    def remove_trait(self, name):
        """ Remove a trait from its name.

        Parameters
        ----------
        name: str (mandatory)
            the trait name to remove.
        """
        # Debug message
        logger.debug("Removing trait '{0}'...".format(name))

        # Call the Traits remove_trait method
        super(Controller, self).remove_trait(name)

        # Remove name from the '_user_traits' without error if it
        # is not present
        self._user_traits.pop(name, None)
        self.user_traits_changed = True

    def export_to_dict(self, exclude_undefined=False,
                       exclude_transient=False,
                       exclude_none=False,
                       exclude_empty=False,
                       dict_class=OrderedDict):
        """ return the controller state to a dictionary, replacing controller
        values in sub-trees to dicts also.

        Parameters
        ----------
        exclude_undefined: bool (optional)
            if set, do not export Undefined values
        exclude_transient: bool (optional)
            if set, do not export values whose trait is marked "transcient"
        exclude_none: bool (optional)
            if set, do not export None values
        exclude_empty: bool (optional)
            if set, do not export empty lists/dicts values
        dict_class: class type (optional, default: soma.sorted_dictionary.OrderedDict)
            use this type of mapping type to represent controllers. It should
            follow the mapping protocol API.
        """
        return controller_to_dict(self, exclude_undefined=exclude_undefined,
                                  exclude_transient=exclude_transient,
                                  exclude_none=exclude_none,
                                  exclude_empty=exclude_empty,
                                  dict_class=dict_class)

    def import_from_dict(self, state_dict, clear=False):
        """ Set Controller variables from a dictionary. When setting values on
        Controller instances (in the Controller sub-tree), replace dictionaries
        by Controller instances appropriately.

        Parameters
        ----------
        state_dict: dict, sorted_dictionary or OrderedDict
            dict containing the variables to set
        clear: bool (optional, default: False)
            if True, older values (in keys not listed in state_dict) will be
            cleared, otherwise they are left in place.
        """
        if clear:
            for trait_name in self.user_traits():
                if trait_name not in state_dict:
                    delattr(self, trait_name)
        for trait_name, value in six.iteritems(state_dict):
            trait = self.trait(trait_name)
            if trait_name == 'protected_parameters' and trait is None:
                HasTraits.add_trait(self, 'protected_parameters',
                                    traits.List(traits.Str(), default=[],
                                                hidden=True))
                trait = self.trait('protected_parameters')
            if trait is None and not isinstance(self, OpenKeyController):
                raise KeyError(
                    "item %s is not a trait in the Controller" % trait_name)
            if isinstance(trait.trait_type, Instance) \
                    and issubclass(trait.trait_type.klass, Controller):
                controller = trait.trait_type.create_default_value(
                    trait.trait_type.klass)
                controller.import_from_dict(value)
            else:
                if value in (None, Undefined):
                    # None / Undefined may be an acceptable value for many
                    # traits types
                    try:
                        setattr(self, trait_name, value)
                    except traits.TraitError:
                        if value is not Undefined:
                            setattr(self, trait_name, Undefined)
                else:
                    # check trait type for conversions
                    tr = self.trait(trait_name)
                    if tr and isinstance(tr.trait_type, Set):
                        setattr(self, trait_name, set(value))
                    elif tr and isinstance(tr.trait_type, Tuple):
                        setattr(self, trait_name, tuple(value))
                    else:
                        setattr(self, trait_name, value)

    def copy(self, with_values=True):
        """ Copy traits definitions to a new Controller object

        Parameters
        ----------
        with_values: bool (optional, default: False)
            if True, traits values will be copied, otherwise the defaut trait
            value will be left in the copy.

        Returns
        -------
        copied: Controller instance
            the returned copy will have the same class as the copied object
            (which may be a derived class from Controller). Traits definitions
            will be copied. Traits values will only be copied if with_values is
            True.
        """
        import copy

        initargs = ()
        if hasattr(self, '__getinitargs__'):
            # if the Controller class is subclassed and needs init parameters
            initargs = self.__getinitargs__()
        copied = self.__class__(*initargs)
        for name, trait in six.iteritems(self.user_traits()):
            copied.add_trait(name, self._clone_trait(trait))
            if with_values:
                setattr(copied, name, getattr(self, name))
        if self.trait('protected_parameters'):
            trait = self.trait('protected_parameters')
            HasTraits.add_trait(copied, 'protected_parameters',
                                self._clone_trait(trait))
            if with_values:
                setattr(copied, 'protected_parameters',
                        getattr(self, 'protected_parameters'))
        return copied

    def reorder_traits(self, traits_list):
        """ Reorder traits in the controller according to a new ordered list.

        If the new list does not contain all user traits, the remaining ones
        will be appended at the end.

        Parameters
        ----------
        traits_list: list
            New list of trait names. This list order will be kept.
        """
        former_traits = set(self._user_traits.sortedKeys)
        for t in traits_list:
            if t not in former_traits:
                raise ValueError("parameter %s is not in Controller traits."
                                 % t)
        new_traits = list(traits_list)
        done = set(new_traits)
        for t in self._user_traits.sortedKeys:
            if t not in done:
                new_traits.append(t)
        self._user_traits.sortedKeys = new_traits

    def protect_parameter(self, param, state=True):
        """ Protect the named parameter.

        Protecting is not a real lock, it just marks the parameter a list of
        "protected" parameters. This is typically used to mark values that have
        been set manually by the user (using the ControllerWidget for instance)
        and that should not be later modified by automatic parameters tweaking
        (such as completion systems).

        Protected parameters are listed in an additional trait,
        "protected_parameters".

        If the "state" parameter is False, then we will unprotect it
        (calling unprotect_parameter())
        """
        if not state:
            return self.unprotect_parameter(param)
        if not self.trait('protected_parameters'):
            # add a 'protected_parameters' trait bypassing the
            # Controller.add_trait mechanism (it will not be a "user_trait")
            HasTraits.add_trait(self, 'protected_parameters',
                                traits.List(traits.Str(), default=[],
                                            hidden=True))
            #self.locked_parameters = []
        protected = set(self.protected_parameters)
        protected.update([param])
        self.protected_parameters = sorted(protected)

    def unprotect_parameter(self, param):
        """ Unprotect the named parameter
        """
        if self.trait('protected_parameters'):
            try:
                self.protected_parameters.remove(param)
            except ValueError:
                pass  # it was not protected.

    def is_parameter_protected(self, param):
        """ Tells whether the given parameter is protected or not
        """
        if not self.trait('protected_parameters'):
            return False
        return param in self.protected_parameters


class OpenKeyController(Controller):

    """ A dictionary-like controller, with "open keys": items may be added
    on the fly, traits are created upon assignation.

    A value trait type should be specified to build the items.

    Usage:

    >>> dict_controller = OpenKeyController(value_trait=traits.Str())
    >>> print(dict_controller.user_traits().keys())
    []
    >>> dict_controller.my_item = 'bubulle'
    >>> print(dict_controller.user_traits().keys())
    ['my_item']
    >>> print(dict_controller.export_to_dict())
    {'my_item': 'bubulle'}
    >>> del dict_controller.my_item
    >>> print(dict_controller.export_to_dict())
    {}
    """
    _reserved_names = set(['trait_added'])

    def __init__(self, value_trait=Any(), *args, **kwargs):
        """ Build an OpenKeyController controller.

        Parameters
        ----------
        value_trait: Trait instance (optional, default: Any())
            trait type to be used when creating traits on the fly
        """
        super(OpenKeyController, self).__init__(*args, **kwargs)
        super(OpenKeyController, self).__setattr__('_value_trait', value_trait)

    def __setattr__(self, name, value):
        if not name.startswith('_') and name not in self.__dict__ \
                and name not in self.traits() \
                and not name in OpenKeyController._reserved_names:
            cloned_trait = self._clone_trait(self._value_trait)
            self.add_trait(name, cloned_trait)
        super(OpenKeyController, self).__setattr__(name, value)

    def __delattr__(self, name):
        if self.trait(name):
            self.remove_trait(name)
        else:
            super(OpenKeyController, self).__delattr__(name)

    def __getinitargs__(self):
        return (self._value_trait, )

    # this specialization does not do anything more than the base class does.
    #def copy(self, with_values=True):
        #""" Copy traits definitions to a new Controller object

        #Parameters
        #----------
        #with_values: bool (optional, default: False)
            #if True, traits values will be copied, otherwise the defaut trait
            #value will be left in the copy.

        #Returns
        #-------
        #copied: Controller instance
            #the returned copy will have the same class as the copied object
            #(which may be a derived class from Controller). Traits definitions
            #will be copied. Traits values will only be copied if with_values is
            #True.
        #"""
        #copied = super(OpenKeyController, self).copy(with_values=with_values)
        #super(OpenKeyController, copied).__setattr__('_value_trait',
                                                     #self._value_trait)
        #return copied


class ControllerTrait(TraitType):

    """ A specialized trait type for Controller values.
    """

    def __init__(self, controller, inner_trait=None, **kwargs):
        """ Build a Controller valued trait.

        Contrarily to Instance(Controller), it ensures better validation when
        assigning values.

        It has the ability to convert values from dictionaries, so the trait
        value can be assigned with a dict, whereas it is actually a Controller.
        This works recursively if the controller contains traits which are also
        controllers.

        Parameters
        ----------
        controller: Controller instance (mandatory)
            default value for trait, and placeholder for allowed traits
        inner_trait: Trait instance (optional)
            if provided, the controller is assumed to be an "open key" type,
            new keys/traits can be added on the fly like in a dictionary, and
            this inner_trait is the trait type used to instantiate new
            traits when new keys are encountered while setting values.
            If inner_trait is not provided, we will look if the controller
            instance is an OpenKeyController, and in such case, take its value
            trait.
        """
        super(ControllerTrait, self).__init__(None, **kwargs)
        self.controller = controller
        self.default_value = controller
        if inner_trait is None and hasattr(controller, '_value_trait'):
            inner_trait = controller._value_trait
        self.inner_trait = inner_trait
        self.handler = self

    def validate(self, object, name, value):
        if isinstance(value, Controller):
            sup_inst = super(ControllerTrait, self)
            if hasattr(sup_inst, 'validate'):
                return sup_inst.validate(value)
            else:
                return value
        if not hasattr(value, 'items'):
            raise TraitError('trait must be a Controller or a mapping type')
        new_value = getattr(object, name).copy(with_values=False)
        if self.inner_trait:
            for key in new_value.user_traits():
                if key not in value:
                    new_value.remove_trait(key)
            for key in value:
                if not self.controller.trait(key):
                    new_value.add_trait(key, self.inner_trait)
        new_value.import_from_dict(value)
        return new_value

    def inner_traits(self):
        if self.inner_trait:
            return (self.inner_trait, )
        return ()


def controller_to_dict(item, exclude_undefined=False,
                       exclude_transient=False,
                       exclude_none=False,
                       exclude_empty=False,
                       dict_class=OrderedDict):
    """
    Convert an item to a Python value where controllers had been converted
    to dictionary. It can recursively convert the values contained in a
    Controller instances or a  dict instances. All other items are returned
    untouched.

    Parameters
    ----------
    exclude_undefined: bool (optional)
        if set, do not export Undefined values
    exclude_transient: bool (optional)
        if set, do not export values whose trait is marked "transcient"
    exclude_none: bool (optional)
        if set, do not export None values
    exclude_empty: bool (optional)
        if set, do not export empty lists/dicts values
    dict_class: class type (optional, default: soma.sorted_dictionary.OrderedDict)
        use this type of mapping type to represent controllers. It should
        follow the mapping protocol API.
    """
    if isinstance(item, Controller):
        result = dict_class()
        for name, trait in six.iteritems(item.user_traits()):
            if exclude_transient and trait.transient:
                continue
            value = getattr(item, name)
            if (exclude_undefined and value is Undefined) \
                or (exclude_none and value is None):
                continue
            if exclude_empty and (value == [] or value == {}):
                continue
            value = controller_to_dict(value,
                                       exclude_undefined=exclude_undefined,
                                       exclude_transient=exclude_transient,
                                       exclude_none=exclude_none,
                                       exclude_empty=exclude_empty,
                                       dict_class=dict_class)
            result[name] = value
        if item.trait('protected_parameters'):
            result['protected_parameters'] = item.protected_parameters
    elif isinstance(item, dict):
        result = dict_class()
        for name, value in six.iteritems(item):
            if (exclude_undefined and value is Undefined) \
                or (exclude_none and value is None):
                continue
            if exclude_empty and (value == [] or value == {}):
                continue
            value = controller_to_dict(value,
                                       exclude_undefined=exclude_undefined,
                                       exclude_transient=exclude_transient,
                                       exclude_none=exclude_none,
                                       exclude_empty=exclude_empty,
                                       dict_class=dict_class)
            result[name] = value
    else:
        result = item

    return result


try:
    import json

    class JsonControllerEncoder(json.JSONEncoder):
        def default(self, obj):
            if obj is Undefined:
                return {'__class__': '<undefined>'}
            if isinstance(obj, traits.TraitSetObject):
                return list(obj) # {'__class__': 'traits.TraitSetObject',
                        #'items': list(obj)}
            if not isinstance(obj, Controller):
                return super(JsonControllerEncoder, self).default(obj)
            d = obj.export_to_dict(exclude_undefined=True,
                exclude_transient=True,
                exclude_none=True,
                exclude_empty=True,
                dict_class=OrderedDict)
            d['__class__'] = obj.__class__.__name__
            return d

    class JsonControllerDecoder(json.JSONDecoder):
        def __init__(self, *args, **kwargs):
            # install a new object_hoook.
            self._old_object_hook = None
            if 'object_hook' in kwargs:
                self._old_object_hook = kwargs['object_hook']
                kwargs = {k: v for k, v in kwargs.items()
                          if k != 'object_hook'}
            super(JsonControllerDecoder, self).__init__(
                *args, object_hook=self.obj_hook, **kwargs)

        def obj_hook(self, obj):
            if self._old_object_hook is not None:
                obj = self._old_object_hook(obj)
            if isinstance(obj, dict) and '__class__' in obj:
                c = obj['__class__']
                if c == '<undefined>':
                    return Undefined
                # Controller objects are decoded as dicts, without the
                # __class__ item, because we cannot rebuild their traits in
                # the general case. They should be converted later by
                # import_from_dict()
                d = {k: v for k, v in obj.items() if k != '__class__'}
                return d
                #controller = Controller()  ## FIXME instantiate __Class__
                #controller.import_from_dict(d)
                #return controller
            return obj

    if type(json._default_encoder) is json.JSONEncoder \
            or json._default_encoder.__class__.__name__ \
                == 'JsonControllerEncoder':
        json._default_encoder = JsonControllerEncoder()
    if type(json._default_decoder) is json.JSONDecoder \
            or json._default_decoder.__class__.__name__ \
                == 'JsonControllerDecoder':
        json._default_decoder = JsonControllerDecoder()

except ImportError:
    pass



# -*- coding: utf-8 -*-
""" Information module, containing version variables, plus some
setuptools-related variables
Attributes
----------
version_major
version_minor
version_micro
version_extra
"""

from __future__ import absolute_import
import sys

version_major = 5
version_minor = 1
version_micro = 1
version_extra = ''
_version_major = version_major
_version_minor = version_minor
_version_micro = version_micro
_version_extra = version_extra

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
__version__ = "%s.%s.%s%s" % (version_major,
                              version_minor,
                              version_micro,
                              version_extra)
CLASSIFIERS = ["Development Status :: 5 - Production/Stable",
               "Environment :: Console",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering",
               "Topic :: Utilities"]

description = 'soma-base'

long_description = """
=========
SOMA-BASE
=========

Miscelaneous all-purpose classes and functions in Python.

"""

# versions for dependencies
SPHINX_MIN_VERSION = '1.0'

# Main setup parameters
NAME = 'soma-base'
PROJECT = 'soma'
ORGANISATION = "CEA"
MAINTAINER = "CEA"
MAINTAINER_EMAIL = ""
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = ""
DOWNLOAD_URL = ""
LICENSE = "CeCILL-B"
CLASSIFIERS = CLASSIFIERS
AUTHOR = "CEA"
AUTHOR_EMAIL = ""
PLATFORMS = "OS Independent"
ISRELEASE = version_extra == ''
VERSION = __version__
PROVIDES = ["soma-base"]
REQUIRES = [
    "six >= 1.13",
    "numpy",
]
EXTRAS_REQUIRE = {
    "doc": ["sphinx>=%s" % SPHINX_MIN_VERSION],
    "crypto": ["pycrypto"],
    "controller": ["traits"],
    "subprocess": ["subprocess32;python_version<'3.2'"],
    "test_utils": ["argparse"],
}



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility classes and functions for debugging.

- author: Yann Cointepas
- organization: NeuroSpin
- license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
from __future__ import print_function

__docformat__ = 'restructuredtext en'

import sys
from pprint import pprint

from soma.undefined import Undefined


def function_call_info(frame=None):
    '''
    Return a dictionary that gives information about a frame corresponding to a function call.
    The directory contains the following items:

    - 'function': name of the function called
    - 'filename': name of the python file containing the function
    - 'lineno': line number executed in 'filename'
    - 'arguments': arguments passed to the function. It is a list containing
      pairs of (argument name, argument value).
    '''
    try:
        if frame is None:
            frame = sys._getframe(1)
        result = {
            'function': frame.f_code.co_name,
            'lineno': frame.f_lineno,
            'filename': frame.f_code.co_filename,
        }
        args = frame.f_code.co_varnames[:frame.f_code.co_argcount]
        result[ 'arguments' ] = \
            [(p, frame.f_locals.get(p, frame.f_globals.get(p, Undefined)))
             for p in args]
    finally:
        del frame
    return result


def stack_calls_info(frame=None):
    '''
    Return a list containing function_call_info(frame) for all frame in the stack.
    '''
    try:
        if frame is None:
            frame = sys._getframe(1)
        result = []
        while frame is not None:
            result.insert(0, function_call_info(frame))
            frame = frame.f_back
        return result
    finally:
        del frame


def print_stack(out=sys.stdout, frame=None):
    '''
    Print information about the stack, including argument passed to functions called.
    '''
    try:
        if frame is None:
            frame = sys._getframe(1)
        for info in stack_calls_info(frame):
            print('File "%(filename)s", line %(lineno)d' % info + ' in '
                  + info['function'], file=out)
            for name, value in info['arguments']:
                out.write('   ' + name + ' = ')
                pprint(value, out, 3)
        out.flush()
    finally:
        del frame



# -*- coding: utf-8 -*-
''' replacement for :func:`ctypes.util.find_library`.
Provides a better version of :func:`find_library`, and allows to patch the
:mod:`ctypes` module to use our version: see :func:`patch_ctypes_find_library`.
'''

from __future__ import absolute_import
import os
import sys
import ctypes.util
import glob

ctypes_find_library = ctypes.util.find_library

def find_library(name):
    ''' :func:`ctypes.util.find_library` is broken on linux at least: it relies
    on ``ldconfig``, which only searches system paths, not user paths nor
    ``LD_LIBRARY_PATH``, or alternatively uses ``gcc``, which is not always
    installed nor configured.

    Here we are looking in ``[[DY]LD_LIBRARY_]PATH`` (depending on the system)
    '''
    def sorted_match(filenames):
        return sorted(filenames)[-1] # probably not the best

    exts = ['.so']
    patterns = [ext + '.*' for ext in exts]
    fname = 'lib' + name
    if sys.platform.startswith('linux'):
        envar = 'LD_LIBRARY_PATH'
    elif sys.platform == 'darwin':
        envar = 'DYLD_LIBRARY_PATH'
        exts = ['.dylib']
        patterns = ['.*' + ext for ext in exts]
    elif sys.platform.startswith('win'):
        envar = 'PATH'
        exts = ['.dll', '.DLL']
        patterns = ['.*' + ext for ext in exts]
    else:
        # other undetermined system (bsd, othe unix...?), assume ELF
        envar = 'LD_LIBRARY_PATH'
    paths = os.environ.get(envar)
    if paths is None:
        # no path: fallback to ctypes
        return ctypes_find_library(name)

    paths = paths.split(os.pathsep)
    names = [fname + ext for ext in exts] + [name + ext for ext in exts]
    patterns = [fname + pattern for pattern in patterns] \
        + [name + pattern for pattern in patterns]
    found = None
    for path in paths:
        for tname in names:
            filename = os.path.join(path, tname)
            if os.path.exists(filename):
                found = filename
                break
        for tname in patterns:
            filenames = glob.glob(os.path.join(path, tname))
            if len(filenames) != 0:
                found = sorted_match(filenames)
                break

    if found is not None:
        return os.path.basename(os.path.realpath(found))

    # not found: fallback to ctypes
    return ctypes_find_library(name)


def patch_ctypes_find_library():
    ''' replace :func:`ctypes.util.find_library` by our version
    '''
    if ctypes.util.find_library is ctypes_find_library:
        ctypes.util.find_library = find_library



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Utility functions for HTML format.

* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import six
import sys

#------------------------------------------------------------------------------
#: mapping of charaters to be escaped for HTML
_htmlEscape = None
_lesserHtmlEscape = None


# ylep 2020-03-24: now that UTF-8 is everywhere, shouldn't we just replace
# HTML-unsafe characters (&<>"') and leave the rest untouched? (i.e. what the
# standard library function html.escape does in Python 3.2 and later).

def htmlEscape(msg):
    """Replace special characters by their correponding html entity.

    All characters that have a corresponding named HTML entity are replaced.

    - returns: *unicode*
    """
    global _htmlEscape
    if _htmlEscape is None:
        _htmlEscape = {
            codepoint: u'&' + name + u';'
            for codepoint, name
            in six.iteritems(six.moves.html_entities.codepoint2name)
        }
    if not isinstance(msg, six.string_types):
        # htmlEscape is sometimes used on non-string types (as print) like
        # tuples or dicts
        msg = str(msg)
    msg = six.ensure_text(msg)
    return msg.translate(_htmlEscape)


def lesserHtmlEscape(msg):
    """Replace special characters by their correponding html entity.

    All characters that have a corresponding named HTML entity are replaced,
    except accented characters commonly used in French text (éàèâêôîûàö) and
    the double-quote character (").

    - returns: *unicode*
    """
    global _lesserHtmlEscape
    if _lesserHtmlEscape is None:
        _lesserHtmlEscape = {
            codepoint: u'&' + name + u';'
            for codepoint, name
            in six.iteritems(six.moves.html_entities.codepoint2name)
            if six.unichr(codepoint) not in (u'"', u'é', u'à', u'è', u'â',
                                             u'ê', u'ô', u'î', u'û', u'ù',
                                             u'ö', )
        }
    msg = six.ensure_text(msg)
    return msg.translate(_lesserHtmlEscape)



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import absolute_import
import logging
from functools import partial
import sys
import six

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.utils.functiontools import SomaPartial
from soma.utils.weak_proxy import weak_proxy
import traits.api as traits
import sip


class EnumControlWidget(object):

    """ Control to select a value from a list.
    """

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control value is correct.

        Parameters
        ----------
        control_instance: QComboBox (mandatory)
            the control widget we want to validate

        Returns
        -------
        out: bool
            always True since the control value is always valid
        """
        return True

    @classmethod
    def check(cls, control_instance):
        """ Check if a controller widget control is filled correctly.

        Parameters
        ----------
        cls: EnumControlWidget (mandatory)
            an EnumControlWidget control
        control_instance: QComboBox (mandatory)
            the control widget we want to validate
        """
        # Hook: function that will be called to check for typo
        # when a 'textEdited' qt signal is emited
        widget_callback = partial(cls.is_valid, weak_proxy(control_instance))

        # Execute manually the first time the control check method
        widget_callback()

        # When a qt 'editTextChanged' signal is emited, check if the new
        # user value is correct
        control_instance.editTextChanged.connect(widget_callback)

    @staticmethod
    def add_callback(callback, control_instance):
        """ Method to add a callback to the control instance when a 'editTextChanged'
        signal is emited.

        Parameters
        ----------
        callback: @function (mandatory)
            the function that will be called when a 'editTextChanged' signal is
            emited.
        control_instance: QComboBox (mandatory)
            the control widget we want to validate
        """
        control_instance.editTextChanged.connect(callback)

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, user_data=None):
        """ Method to create the widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: str (mandatory)
            the default control value, here the enum values
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: QComboBox,
            associated label: QLabel)
        """
        # Create the widget that will be used to select a value
        widget = QtGui.QComboBox(parent)

        # Save the possible choices
        widget._choices = trait.handler.values

        # Add a parameter to tell us if the widget is optional
        widget.optional = trait.optional

        # Set the enum list items to the widget
        for item in widget._choices:
            widget.addItem(six.text_type(item))

        # Select the default value
        # If the default value is not in the enum list, pick the first item
        # of the enum list
        if control_value not in widget._choices:
            widget.setCurrentIndex(0)
        else:
            widget.setCurrentIndex(widget._choices.index(control_value))

        # Create the label associated with the enum widget
        control_label = trait.label
        if control_label is None:
            control_label = control_name
        if label_class is None:
            label_class = QtGui.QLabel
        if control_label is not None:
            label = label_class(control_label, parent)
        else:
            label = None

        return (widget, label)

    @staticmethod
    def update_controller(controller_widget, control_name,
                          control_instance, *args, **kwargs):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: StrControlWidget (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        if EnumControlWidget.is_valid(control_instance):
            new_trait_value = control_instance._choices[
                control_instance.currentIndex()]
            # value is manually modified: protect it
            if getattr(controller_widget.controller, control_name) \
                    != new_trait_value:
                controller_widget.controller.protect_parameter(control_name)
            setattr(controller_widget.controller, control_name,
                    new_trait_value)
            logger.debug(
                "'EnumControlWidget' associated controller trait '{0}' "
                " has been updated with value '{1}'.".format(
                    control_name, new_trait_value))

    @staticmethod
    def update_controller_widget(controller_widget, control_name,
                                 control_instance):
        """ Update one element of the controller widget.

        At the end the controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: StrControlWidget (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """

        # Get the controller trait value
        try:
            test = control_instance.setCurrentIndex
        except ReferenceError:
            # widget deleted in the meantime
            return

        if sip.isdeleted(control_instance.__init__.__self__):
            EnumControlWidget.disconnect(controller_widget, control_name,
                                         control_instance)
            return

        new_controller_value = getattr(
            controller_widget.controller, control_name, None)

        # If the controller value is not empty, update the controller widget
        # associated control
        if new_controller_value not in (None, traits.Undefined):
            control_instance.setCurrentIndex(
                control_instance._choices.index(new_controller_value))
        logger.debug("'EnumControlWidget' has been updated with value "
                     "'{0}'.".format(new_controller_value))

    @classmethod
    def connect(cls, controller_widget, control_name, control_instance):
        """ Connect an 'Enum' controller trait and an 'EnumControlWidget'
        controller widget control.

        Parameters
        ----------
        cls: EnumControlWidget (mandatory)
            an EnumControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QComboBox (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Update one element of the controller.
        # Hook: function that will be called to update a specific
        # controller trait when an 'activated' qt signal is emited
        widget_hook = partial(cls.update_controller,
                              weak_proxy(controller_widget),
                              control_name, weak_proxy(control_instance))

        # When a qt 'activated' signal is emited, update the
        # 'control_name' controller trait value
        control_instance.activated.connect(widget_hook)

        # Update one element of the controller widget.
        # Hook: function that will be called to update the specific widget
        # when a trait event is detected.
        controller_hook = SomaPartial(
            cls.update_controller_widget, weak_proxy(controller_widget),
            control_name, weak_proxy(control_instance))

        # When the 'control_name' controller trait value is modified, update
        # the corresponding control
        controller_widget.controller.on_trait_change(
            controller_hook, name=control_name, dispatch='ui')

        # Store the trait - control connection we just build
        control_instance._controller_connections = (
            widget_hook, controller_hook)
        logger.debug("Add 'Enum' connection: {0}.".format(
            control_instance._controller_connections))

    @staticmethod
    def disconnect(controller_widget, control_name, control_instance):
        """ Disconnect an 'Enum' controller trait and an 'EnumControlWidget'
        controller widget control.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QComboBox (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Get the stored widget and controller hooks
        (widget_hook,
         controller_hook) = control_instance._controller_connections

        # Remove the controller hook from the 'control_name' trait
        controller_widget.controller.on_trait_change(
            controller_hook, name=control_name, remove=True)

        # Remove the widget hook associated with the qt 'activated'
        # signal
        if not sip.isdeleted(control_instance.__init__.__self__):
            control_instance.activated.disconnect(widget_hook)

        # Delete the trait - control connection we just remove
        del control_instance._controller_connections



# -*- coding: utf-8 -*-

from __future__ import absolute_import
import re
from inspect import isfunction, isclass, ismethod
from weakref import WeakKeyDictionary

from soma.singleton import Singleton


class GlobalNaming(Singleton):
    """ No doc ?
    """

    def __singleton_init__(self):
        super(GlobalNaming, self).__singleton_init__()
        self._global_name_re = re.compile(
            r'(([A-Za-z][A-Za-z0-9_.]*)\.([A-Za-z0-9_]+))(\(\))?((\.)([A-Za-z][A-Za-z0-9_.]+))?')
        self._names = WeakKeyDictionary()

    def parse_global_name(self, global_name):
        """
        """
        m = self._global_name_re.match(global_name)
        if m:
            # return module, name, args, attributes
            return m.group(2), m.group(3), (() if m.group(4) else None), m.group(7)
        raise ValueError('Invalid global name syntax: ' + global_name)

    def get_object(self, global_name):
        """
        """
        module, name, args, attributes = self.parse_global_name(global_name)
        module = __import__(module, fromlist=[name], level=0)
        try:
            value = getattr(module, name)
        except AttributeError as e:
            raise ImportError(str(e))
        if args is not None:
            value = value(*args)
        if attributes:
            for a in attributes.split('.'):
                value = getattr(value, a)
            self._names[value] = global_name
        return value

    def get_name(self, obj):
        """
        """
        result = self._names.get(obj)
        if result is not None:
            return result
        if isfunction(obj) or isclass(obj):
            return obj.__module__ + '.' + obj.__name__
        elif ismethod(obj):
            if obj.__self__ is not None:
                return obj.__class__.__module__ + '.' + obj.__class__.__name__ + '().' + obj.__func__.__name__
        elif isinstance(obj, object):
            return obj.__class__.__module__ + '.' + obj.__class__.__name__ + '()'
        raise ValueError('Cannot find global name for %s' % repr(obj))


def get_object(global_name):
    """
    """
    return GlobalNaming().get_object(global_name)


def get_name(obj):
    """
    """
    return GlobalNaming().get_name(obj)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
import json
import os
import collections
import datetime
import glob
import six
import sys
from soma.sorted_dictionary import SortedDictionary
from soma.controller import Controller
try:
    from traits.api import HasTraits, Str, Enum, Directory, File
except ImportError:
    from enthought.traits.api import HasTraits, Str, Enum, Directory, File
from soma.application import Application


class Study(Controller):
    _instance = None
    """Class to write and save informations about process in the json"""

    def __init__(self):
        super(Study, self).__init__()
        HasTraits.__init__(self)
        # Find foms available
        foms = Application().fom_manager.find_foms()
        foms.insert(0, ' ')
        self.add_trait('input_directory', Directory)
        # self.add_trait('input_directory',Directory('/nfs/neurospin/cati/cati_shared'))
        self.add_trait('input_fom', Enum(foms))
        self.add_trait('output_directory', Directory)
        self.add_trait('output_fom', Enum(foms))
        self.add_trait('shared_directory', Directory)
        self.add_trait('spm_directory', Directory)
        self.add_trait('format_image', Str)
        self.add_trait('format_mesh', Str)
        self.add_trait('process', Enum(
            '   ',
            'morphologist.process.morphologist_simplified.SimplifiedMorphologist',
            'morphologist.process.morphologist.Morphologist',
            'morpho.morphologist.morphologist',
            'morphologist.process.customized.morphologist.CustomMorphologist'))
        # self.process_specific=None
        self.compteur_run_process = {}
        self.runs = collections.OrderedDict()

    # def _input_fom_changed(self, old, new):
        # print 'input fom changed'
        # Get list format
        # self.add_trait(œ'coucou',Str)

    @staticmethod
    def get_instance():
        if Study._instance is None:
            Study._instance = Study()
            return Study._instance
        else:
            return Study._instance

    """Save on json with OrderedDict"""

    def save(self):
        self.name_study = str(self.output_directory.split(os.sep)[-1])
        self.dico = collections.OrderedDict([('name_study', self.name_study), ('input_directory', self.input_directory), ('input_fom', self.input_fom), ('output_directory', self.output_directory), (
            'output_fom', self.output_fom), ('shared_directory', self.shared_directory), ('spm_directory', self.spm_directory), ('format_image', self.format_image), ('format_mesh', self.format_mesh), ('process', self.process)])
        json_string = json.dumps(self.dico, indent=4, separators=(',', ': '))
        with open(os.path.join(self.output_directory, self.name_study + '.json'), 'w') as f:
            f.write(six.text_type(json_string))

    """Load and put on self.__dict__ OrderedDict"""

    def load(self, name_json):
        try:
            with open(name_json, 'r') as json_data:
                self.__dict__ = json.load(
                    json_data, object_pairs_hook=collections.OrderedDict)
            for element in self.__dict__:
                setattr(self, element, self.__dict__[element])

        # No file to load
        except IOError:
            pass

    #"""Get number of run process and iterate"""
    # def inc_nb_run_process(self,process_name):
        # print 'process_name',process_name
        # if process_name in self.compteur_run_process:
            # valeur=self.compteur_run_process[process_name]
            # print 'valeur',valeur
            # self.compteur_run_process[process_name]=valeur+1
        # else:
            # self.compteur_run_process[process_name]=1
    def save_run(self, attributes, process):
        # Create date directory
        date = datetime.datetime.now()
        date_directory = str(date.day) + '_' + str(
            date.month) + '_' + str(date.year)
        directory = os.path.join(self.output_directory, date_directory)
        if not os.path.exists(directory):
            name_run = process.name + str(1)
            os.makedirs(directory)
            self.compteur_run_process[process.name] = 1

        else:
            list_json = glob.glob(directory + os.sep + '*.json')
            inc = 1
            number = len(list_json) + inc
            name_run = process.name + str(number)
            while os.path.exists(os.path.join(directory, name_run + '.json')) is True:
                inc = inc + 1
                number = len(list_json) + inc
                name_run = process.name + str(number)
            self.compteur_run_process[process.name] = number

        run = collections.OrderedDict()
        run['process_name'] = 'morphologist.process.morphologist_simplified.SimplifiedMorphologist'
        # self.inc_nb_run_process(process.name )
        run['attributes'] = {}
        for key in attributes:
            run['attributes'][key] = attributes[key]
        run['parameters'] = collections.OrderedDict()
        run['input'] = collections.OrderedDict()
        run['output'] = collections.OrderedDict()
        # dicti_sorted=SortedDictionary(*[(key,a[key]) for key in a])

        for name, trait in six.iteritems(process.user_traits()):
            if trait.is_trait_type(File) is False:
                run['parameters'][name] = getattr(process, name)
            elif trait.output is False:
                run['input'][name] = getattr(process, name)
            elif trait.output is True:
                run['output'][name] = getattr(process, name)

        dico = collections.OrderedDict([('name_study', self.name_study), ('input_directory', self.input_directory), ('input_fom', self.input_fom), ('output_directory', self.output_directory), (
            'output_fom', self.output_fom), ('shared_directory', self.shared_directory), ('spm_directory', self.spm_directory), ('format_image', self.format_image), ('format_mesh', self.format_mesh), ('run', run)])
        json_string = json.dumps(dico, indent=4, separators=(',', ': '))

        with open(os.path.join(directory, name_run + '.json'), 'w') as f:
            f.write(six.text_type(json_string))

        # To incremente compteur runs
        self.save()



# -*- coding: utf-8 -*-
#
# SOMA - Copyright (C) CEA, 2015
# Distributed under the terms of the CeCILL-B license, as published by
# the CEA-CNRS-INRIA. Refer to the LICENSE file or to
# http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html
# for details.
#

# System import
from __future__ import print_function
from __future__ import absolute_import
import os
import logging
from functools import partial
import six
from six.moves import range

# Define the logger
logger = logging.getLogger(__name__)

# Soma import
from soma.qt_gui import qt_backend
from soma.qt_gui.qt_backend import QtGui, QtCore
from soma.utils.functiontools import SomaPartial
from soma.controller import trait_ids
from soma.controller import Controller
from soma.qt_gui.controller_widget import ControllerWidget, get_ref, weak_proxy
import traits.api as traits
import json
import csv
import sys
import sip

from six.moves import cStringIO as StringIO

# Qt import
try:
    _fromUtf8 = QtCore.QString.fromUtf8
except AttributeError:
    _fromUtf8 = lambda s: s

QtCore.QResource.registerResource(os.path.join(os.path.dirname(
    os.path.dirname(__file__)), 'resources', 'widgets_icons.rcc'))


class ListController(Controller):

    """ Dummy list controller to simplify the creation of a list widget
    """
    pass


class ListControlWidget(object):

    """ Control to enter a list of items.
    """

    #
    # Public members
    #

    @staticmethod
    def is_valid(control_instance, *args, **kwargs):
        """ Method to check if the new control values are correct.

        If the new list controls values are not correct, the backroung
        color of each control in the list will be red.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the control widget we want to validate

        Returns
        -------
        valid: bool
            True if the control values are valid,
            False otherwise
        """
        # Initilaized the output
        valid = True

        # If the trait is optional, the control is valid
        if control_instance.trait.optional is True:
            return valid

        # Go through all the controller widget controls
        controller_widget = control_instance.controller_widget
        for control_name, control_groups \
                in six.iteritems(controller_widget._controls):

            if not control_groups:
                continue
            # Unpack the control item
            trait, control_class, control_instance, control_label \
                = next(iter(control_groups.values()))

            # Call the current control specific check method
            valid = control_class.is_valid(control_instance)

            # Stop checking if a wrong control has been found
            if not valid:
                break

        return valid

    @classmethod
    def check(cls, control_instance):
        """ Check if a controller widget list control is filled correctly.

        Parameters
        ----------
        cls: ListControlWidget (mandatory)
            a ListControlWidget control
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def add_callback(callback, control_instance):
        """ Method to add a callback to the control instance when the list
        trait is modified

        Parameters
        ----------
        callback: @function (mandatory)
            the function that will be called when a 'textChanged' signal is
            emited.
        control_instance: QFrame (mandatory)
            the control widget we want to validate
        """
        pass

    @staticmethod
    def create_widget(parent, control_name, control_value, trait,
                      label_class=None, max_items=0, user_data=None):
        """ Method to create the list widget.

        Parameters
        ----------
        parent: QWidget (mandatory)
            the parent widget
        control_name: str (mandatory)
            the name of the control we want to create
        control_value: list of items (mandatory)
            the default control value
        trait: Tait (mandatory)
            the trait associated to the control
        label_class: Qt widget class (optional, default: None)
            the label widget will be an instance of this class. Its constructor
            will be called using 2 arguments: the label string and the parent
            widget.
        max_items: int (optional)
            display at most this number of items. Defaults to 0: no limit.

        Returns
        -------
        out: 2-uplet
            a two element tuple of the form (control widget: ,
            associated labels: (a label QLabel, the tools QWidget))
        """
        # Get the inner trait: expect only one inner trait
        # note: trait.inner_traits might be a method (ListInt) or a tuple
        # (List), whereas trait.handler.inner_trait is always a method
        if len(trait.handler.inner_traits()) == 1:
            inner_trait = trait.handler.inner_traits()[0]
        elif len(trait.handler.inner_traits()) == 0:
            # maybe a generic list, or a compount trait
            if hasattr(trait.handler, 'handlers') \
                    and len(trait.handler.handlers) > 0 \
                    and hasattr(trait.handler.handlers[0], 'inner_traits') \
                    and len(trait.handler.handlers[0].inner_traits()) > 0:
                inner_trait = trait.handler.handlers[0].inner_traits()[0]
            else:
                # can't determine type, fallback to string
                inner_trait = traits.Str()
        else:
            raise Exception(
                "Expect only one inner trait in List control. Trait '{0}' "
                "inner trait is '{1}'.".format(control_name,
                                               trait.handler.inner_traits()))

        if control_value is traits.Undefined:
            control_value = []
        elif not isinstance(control_value, (list, tuple)):
            # in nipype MultiPath, single values are not in a list
            control_value = [control_value]
            
        # Create the list widget: a frame
        parent = get_ref(parent)
        frame = QtGui.QFrame(parent=parent)
        #frame.setFrameShape(QtGui.QFrame.StyledPanel)
        frame.setFrameShape(QtGui.QFrame.NoFrame)

        # Create tools to interact with the list widget: expand or collapse -
        # add a list item - remove a list item
        tool_widget = QtGui.QWidget(parent)
        layout = QtGui.QHBoxLayout()
        layout.addStretch(1)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        tool_widget.setLayout(layout)
        # Create the tool buttons
        resize_button = QtGui.QToolButton()
        add_button = QtGui.QToolButton()
        delete_button = QtGui.QToolButton()
        layout.addWidget(resize_button)
        layout.addWidget(add_button)
        layout.addWidget(delete_button)
        # Set the tool icons
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/add")),
                       QtGui.QIcon.Normal, QtGui.QIcon.Off)
        add_button.setIcon(icon)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/delete")),
                       QtGui.QIcon.Normal, QtGui.QIcon.Off)
        delete_button.setIcon(icon)
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/nav_down")),
            QtGui.QIcon.Normal, QtGui.QIcon.Off)
        resize_button.setIcon(icon)
        resize_button.setFixedSize(30, 22)
        add_button.setFixedSize(40, 22)
        delete_button.setFixedSize(40, 22)

        menu = QtGui.QMenu()
        menu.addAction('Enter list',
                       partial(ListControlWidget.enter_list,
                               weak_proxy(parent), control_name,
                               weak_proxy(frame)))
        menu.addAction('Load list',
                       partial(ListControlWidget.load_list,
                               weak_proxy(parent), control_name,
                               weak_proxy(frame)))
        if isinstance(inner_trait.trait_type, traits.File) \
                or isinstance(inner_trait.trait_type, traits.Directory):
            menu.addAction('Select files',
                           partial(ListControlWidget.select_files,
                                   weak_proxy(parent),
                                   control_name, weak_proxy(frame)))
        add_button.setMenu(menu)

        menu = QtGui.QMenu()
        menu.addAction('Clear all',
                       partial(ListControlWidget.clear_all,
                               weak_proxy(parent), control_name,
                               weak_proxy(frame), trait.trait_type.minlen))
        delete_button.setMenu(menu)

        # Create a new controller that contains length 'control_value' inner
        # trait elements
        controller = ListController()

        if inner_trait.groups:
            del inner_trait.groups

        n = max_items
        if n == 0:
            n = len(control_value)

        for cnt, inner_control_values in enumerate(control_value[:n]):
            controller.add_trait(str(cnt), inner_trait)
            #if inner_trait.groups:
                #del trait(str(cnt)).groups
            setattr(controller, str(cnt), inner_control_values)

        # Create the associated controller widget
        controller_widget = ControllerWidget(
            controller, parent=frame, live=True,
            override_control_types=parent._defined_controls,
            user_data=user_data)
        controller_widget.setObjectName('inner_controller')
        controller_widget.setStyleSheet(
            'ControllerWidget#inner_controller { padding: 0px; }')

        # Store some parameters in the list widget
        frame.inner_trait = inner_trait
        frame.trait = trait
        frame.controller = controller
        frame.controller_widget = controller_widget
        frame.connected = False
        frame.max_items = max_items

        # Add the list controller widget to the list widget
        frame.setLayout(controller_widget.layout())
        frame.layout().setContentsMargins(0, 0, 0, 0)
        frame.setObjectName('inner_frame')
        frame.setStyleSheet('QFrame#inner_frame { padding: 0px; }')

        # Set some callback on the list control tools
        # Resize callback
        resize_hook = partial(
            ListControlWidget.expand_or_collapse, weak_proxy(frame),
            weak_proxy(resize_button))
        resize_button.clicked.connect(resize_hook)
        # Add list item callback
        add_hook = partial(
            ListControlWidget.add_list_item, weak_proxy(parent),
            control_name, weak_proxy(frame))
        add_button.clicked.connect(add_hook)
        # Delete list item callback
        delete_hook = partial(
            ListControlWidget.delete_list_item, weak_proxy(parent),
            control_name, weak_proxy(frame))
        delete_button.clicked.connect(delete_hook)

        # Create the label associated with the list widget
        control_label = trait.label
        if control_label is None:
            control_label = control_name
        if label_class is None:
            label_class = QtGui.QLabel
        if control_label is not None:
            label = label_class(control_label, parent)
        else:
            label = None

        return (frame, (label, tool_widget))

    @staticmethod
    def update_controller(controller_widget, control_name, control_instance,
                          *args, **kwarg):
        """ Update one element of the controller.

        At the end the controller trait value with the name 'control_name'
        will match the controller widget user parameters defined in
        'control_instance'.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Get the list widget inner controller values
        new_trait_value = [
            getattr(control_instance.controller, str(i))
            for i in range(len(control_instance.controller.user_traits()))]

        #if not hasattr(control_instance, 'max_items'):
            #print('no max_items in:', control_instance)
        #else:
        if control_instance.max_items != 0 \
                and len(new_trait_value) == control_instance.max_items:
            old_value = getattr(controller_widget.controller, control_name)
            new_trait_value += old_value[control_instance.max_items:]

        updating = getattr(controller_widget, '_updating', False)
        controller_widget._updating = True

        protected = controller_widget.controller.is_parameter_protected(
            control_name)
        # value is manually modified: protect it
        if getattr(controller_widget.controller, control_name) \
                != new_trait_value:
            controller_widget.controller.protect_parameter(control_name)
        # Update the 'control_name' parent controller value
        try:
            setattr(controller_widget.controller, control_name,
                    new_trait_value)
        except Exception as e:
            print(e, file=sys.stderr)
            if not protected:
                controller_widget.controller.unprotect_parameter(control_name)

        controller_widget._updating = updating
        logger.debug(
            "'ListControlWidget' associated controller trait '{0}' has "
            "been updated with value '{1}'.".format(
                control_name, new_trait_value))

    @staticmethod
    def validate_all_values(controller_widget, control_instance):
        '''Performs recursively update_controller() on list elements to
        make the Controller instance values match values in widgets.
        '''
        for k, groups in six.iteritems(controller_widget._controls):
            for g, ctrl in six.iteritems(groups):
                ctrl[1].update_controller(controller_widget, k,
                                          control_instance,
                                          reset_invalid_value=True)
                if ctrl[1] is ListControlWidget:
                    frame = ctrl[2]
                    sub_cw = frame.controller_widget
                    for sub_k, sub_groups in six.iteritems(sub_cw._controls):
                        for sub_g, sub_ctrl in six.iteritems(sub_groups):
                            sub_ctrl[1].update_controller(
                                sub_cw, sub_k, sub_ctrl[2],
                                reset_invalid_value=True)


    @classmethod
    def update_controller_widget(cls, controller_widget, control_name,
                                 control_instance):
        """ Update one element of the list controller widget.

        At the end the list controller widget user editable parameter with the
        name 'control_name' will match the controller trait value with the same
        name.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # there are 2 Controller instances here:
        # * controller_widget.controller is the "official" edited controller,
        #   which contains a trait "control_name" with a list value
        # * control_instance.controller is a proxy Controller built here
        #   that represents the list items: elements keys are indexes.
        #   This controller has another widget (for list items):
        #   control_instance.controller_widget.
        try:
            was_connected = control_instance.connected
        except ReferenceError:
            # widget deleted in the meantime
            return
        if getattr(controller_widget, '_updating', False):
            return
        controller_widget._updating = True

        if sip.isdeleted(control_instance.__init__.__self__):
            ListControlWidget.disconnect(controller_widget, control_name,
                                         control_instance)
            return

        # One callback has not been removed properly
        if control_name in controller_widget.controller.user_traits():

            # Get the list widget current connection status
            was_connected = control_instance.connected

            # Disconnect the list controller and the inner list controller
            cls.disconnect(controller_widget, control_name, control_instance)
            control_instance.controller_widget.disconnect()

            # Get the 'control_name' list value from the top list controller
            trait_value = getattr(controller_widget.controller, control_name)
            if trait_value in (None, traits.Undefined):
                trait_value = []
            elif not isinstance(trait_value, (tuple, list)):
                # nipype MultiPath may have one item outside of a list
                trait_value = [trait_value]

            # Get the number of list elements in the controller associated
            # with the current list control
            len_widget = len(control_instance.controller.user_traits())

            # Parameter that is True if a user trait associated with the
            # current list control has changed
            user_traits_changed = False

            # Special case: some traits have been deleted to the top controller
            if len(trait_value) < len_widget:

                # Need to remove to the inner list controller some traits
                for i in range(len(trait_value), len_widget):
                    control_instance.controller.remove_trait(str(i))

                # Notify that some traits of the inner list controller have
                # been deleted
                user_traits_changed = True

            # Special case: some new traits have been added to the top
            # controller
            elif len(trait_value) > len_widget \
                    and (control_instance.max_items == 0
                         or len_widget < control_instance.max_items):

                # Need to add to the inner list controller some traits
                # with type 'inner_trait'
                for i in range(len_widget, len(trait_value)):
                    control_instance.controller.add_trait(
                        str(i), control_instance.inner_trait)

                # Notify that some traits of the inner list controller
                # have been added
                user_traits_changed = True

            # Update the controller associated with the current control
            n = len(trait_value)
            if control_instance.max_items != 0 \
                    and control_instance.max_items < n:
                n = control_instance.max_items
            for i in range(n):
                setattr(control_instance.controller, str(i), trait_value[i])

            # Connect the inner list controller
            control_instance.controller_widget.connect()

            # Emit the 'user_traits_changed' signal if necessary
            if user_traits_changed:
                control_instance.controller.user_traits_changed = True

                logger.debug(
                    "'ListControlWidget' inner controller has been updated:"
                    "old size '{0}', new size '{1}'.".format(
                        len_widget, len(trait_value)))

            # Restore the previous list controller connection status
            if was_connected:
                cls.connect(controller_widget, control_name, control_instance)

        else:
            logger.error("oups: control_name %s not in list controller "
              "traits: %s"
              % (control_name,
                 repr(list(
                    controller_widget.controller.user_traits().keys()))))
            # print cls, controller_widget, control_name, control_instance
            # print control_instance.controller
            # print control_instance.controller.user_traits()

        controller_widget._updating = False

    @classmethod
    def connect(cls, controller_widget, control_name, control_instance):
        """ Connect a 'List' controller trait and a 'ListControlWidget'
        controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if not control_instance.connected:

            # Update the list item when one of his associated controller trait
            # changed.
            # Hook: function that will be called to update the controller
            # associated with a list widget when a list widget inner controller
            # trait is modified.
            list_controller_hook = SomaPartial(
                cls.update_controller, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # Go through all list widget inner controller user traits
            for trait_name in control_instance.controller.user_traits():

                # And add the callback on each user trait
                control_instance.controller.on_trait_change(
                    list_controller_hook, trait_name, dispatch='ui')
                logger.debug("Item '{0}' of a 'ListControlWidget', add "
                             "a callback on inner controller trait "
                             "'{0}'.".format(control_name, trait_name))

            # Update the list controller widget.
            # Hook: function that will be called to update the specific widget
            # when a trait event is detected on the list controller.
            controller_hook = SomaPartial(
                cls.update_controller_widget, weak_proxy(controller_widget),
                control_name, weak_proxy(control_instance))

            # When the 'control_name' controller trait value is modified,
            # update the corresponding control
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, dispatch='ui')

            # Update the list connection status
            control_instance._controller_connections = (
                list_controller_hook, controller_hook)
            logger.debug("Add 'List' connection: {0}.".format(
                control_instance._controller_connections))

            # Connect also all list items
            inner_controls = control_instance.controller_widget._controls
            for (inner_control_name,
                 inner_control_groups) in six.iteritems(inner_controls):
                for group, inner_control \
                        in six.iteritems(inner_control_groups):

                    # Unpack the control item
                    inner_control_instance = inner_control[2]
                    inner_control_class = inner_control[1]

                    # Call the inner control connect method
                    inner_control_class.connect(
                        control_instance.controller_widget, inner_control_name,
                        inner_control_instance)

            # Update the list control connection status
            control_instance.connected = True

    @staticmethod
    def disconnect(controller_widget, control_name, control_instance):
        """ Disconnect a 'List' controller trait and a 'ListControlWidget'
        controller widget control.

        Parameters
        ----------
        cls: StrControlWidget (mandatory)
            a StrControlWidget control
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str (mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Check if the control is connected
        if control_instance.connected:

            # Get the stored widget and controller hooks
            (list_controller_hook,
             controller_hook) = control_instance._controller_connections

            # Remove the controller hook from the 'control_name' trait
            controller_widget.controller.on_trait_change(
                controller_hook, control_name, remove=True)

            # Remove the list controller hook associated with the controller
            # traits
            for trait_name in control_instance.controller.user_traits():
                control_instance.controller.on_trait_change(
                    list_controller_hook, trait_name, remove=True)

            # Delete the trait - control connection we just remove
            del control_instance._controller_connections

            # Disconnect also all list items
            inner_controls = control_instance.controller_widget._controls
            for (inner_control_name,
                 inner_control_groups) in six.iteritems(inner_controls):
                for group, inner_control \
                        in six.iteritems(inner_control_groups):

                    # Unpack the control item
                    inner_control_instance = inner_control[2]
                    inner_control_class = inner_control[1]

                    # Call the inner control disconnect method
                    inner_control_class.disconnect(
                        control_instance.controller_widget, inner_control_name,
                        inner_control_instance)

            # Update the list control connection status
            control_instance.connected = False

    #
    # Callbacks
    #

    @staticmethod
    def add_list_item(controller_widget, control_name, control_instance):
        """ Append one element in the list controller widget.

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Get the number of traits associated with the current list control
        # controller
        nb_of_traits = len(control_instance.controller.user_traits())
        if control_instance.max_items != 0 \
                and nb_of_traits >= control_instance.max_items:
            # don't display more.
            return
        trait_name = str(nb_of_traits)

        # Add the new trait to the inner list controller
        control_instance.controller.add_trait(
            trait_name, control_instance.inner_trait)
        control_instance._updating = True
        setattr(control_instance.controller, trait_name,
                control_instance.inner_trait.default)
        control_instance._updating = False

        # Update the list controller
        if hasattr(control_instance, '_controller_connections'):
            control_instance._controller_connections[0]()

        # control_instance.controller_widget.update_controller_widget()
        logger.debug("Add 'ListControlWidget' '{0}' new trait "
                     "callback.".format(trait_name))

    @staticmethod
    def delete_list_item(controller_widget, control_name, control_instance):
        """ Delete the last control element

        Parameters
        ----------
        controller_widget: ControllerWidget (mandatory)
            a controller widget that contains the controller we want to update
        control_name: str(mandatory)
            the name of the controller widget control we want to synchronize
            with the controller
        control_instance: QFrame (mandatory)
            the instance of the controller widget control we want to
            synchronize with the controller
        """
        # Delete the last inserted control
        #print('delete_list_item', controller_widget, control_name, control_instance)

        # inner controller for list items
        controller = control_instance.controller
        keys = controller.user_traits().keys()
        keys = sorted([int(k) for k in keys])
        if not keys:
            print('no element to remove')
            return

        last_key = str(keys[-1])
        controller.remove_trait(last_key)

        # Update the list controller
        if hasattr(control_instance, '_controller_connections'):
            control_instance._controller_connections[0]()
        # notification will update the controller GUI.

    @staticmethod
    def expand_or_collapse(control_instance, resize_button):
        """ Callback to expand or collapse a 'ListControlWidget'.

        Parameters
        ----------
        control_instance: QFrame (mandatory)
            the list widget item
        resize_button: QToolButton
            the signal sender
        """
        # Change the icon depending on the button status
        icon = QtGui.QIcon()

        # Hide the control
        if control_instance.isVisible():
            control_instance.hide()
            icon.addPixmap(
                QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/nav_right")),
                QtGui.QIcon.Normal, QtGui.QIcon.Off)

        # Show the control
        else:
            control_instance.show()
            icon.addPixmap(
                QtGui.QPixmap(_fromUtf8(":/soma_widgets_icons/nav_down")),
                QtGui.QIcon.Normal, QtGui.QIcon.Off)

        # Set the new button icon
        resize_button.setIcon(icon)

    @staticmethod
    def enter_list(controller_widget, control_name, control_instance):
        controller_widget = get_ref(controller_widget)
        widget = ListValuesEditor(controller_widget, controller_widget,
                                  control_name)
        done = False
        while not done:
            if widget.exec_():
                parent_controller = controller_widget.controller
                elem_trait \
                    = parent_controller.trait(control_name).inner_traits[0]
                value = ListControlWidget.parse_list(
                    widget.textedit.toPlainText(),
                    widget.format_c.currentText(),
                    widget.separator_c.currentText(), elem_trait)
                if value is not None:
                    try:
                        setattr(parent_controller, control_name, value)
                    except Exception as e:
                        print(e)
                    done = True
                else:
                    r = QtGui.QMessageBox.warning(
                        controller_widget, 'Parsing error',
                        'Could not parse the text input',
                        QtGui.QMessageBox.Retry | QtGui.QMessageBox.Abort,
                        QtGui.QMessageBox.Retry)
                    if r == QtGui.QMessageBox.Abort:
                        done = True
            else:
                done = True
    @staticmethod
    def parse_list(text, format, separator, elem_trait):
        if format == 'JSON':
            formats = [format, 'CSV']
        elif format == 'CSV':
            formats = [format, 'JSON']
        else:
            formats = ['JSON', 'CSV']
        for format in formats:
            if format == 'JSON':
                try:
                    parsed = json.loads(text)
                    return parsed
                except Exception:
                    pass
            elif format == 'CSV':
                try:
                    reader = csv.reader(
                        text.split('\n'), delimiter=str(separator),
                        quotechar='"', quoting=csv.QUOTE_MINIMAL,
                        skipinitialspace=True)
                    ctrait = traits.TraitCastType(
                        type(elem_trait.default))
                    c = traits.HasTraits()
                    c.add_trait('x', ctrait)
                    parsed = []
                    for x in reader:
                        if isinstance(x, list):
                            for y in x:
                                c.x = y
                                parsed.append(c.x)
                        else:
                            c.x = x
                            parsed.append(c.x)
                    return parsed
                except Exception:
                    pass
        # could not parse
        return None


    @staticmethod
    def load_list(controller_widget, control_name, control_instance):
        controller_widget = get_ref(controller_widget)
        control_instance = get_ref(control_instance)

        # get widget via a __self__ in a method, because control_instance may
        # be a weakproxy.
        widget = control_instance.__repr__.__self__

        fname = qt_backend.getOpenFileName(
            widget, "Open file", "", "", None,
            QtGui.QFileDialog.DontUseNativeDialog)
        if fname:
            parent_controller = controller_widget.controller
            elem_trait = parent_controller.trait(control_name).inner_traits[0]
            text = open(fname).read()
            if fname.endswith('.csv'):
                format = 'CSV'
            else:
                format = 'JSON'
            value = ListControlWidget.parse_list(text, format, ',', elem_trait)
            if value is not None:
                setattr(parent_controller, control_name, value)
            else:
                QtGui.QMessageBox.warning(
                    controller_widget, 'Parsing error',
                    'Could not parse the input file',
                    QtGui.QMessageBox.Cancel, QtGui.QMessageBox.Cancel)

    @staticmethod
    def select_files(controller_widget, control_name, control_instance):
        control_instance = get_ref(control_instance)
        parent_controller = controller_widget.controller
        elem_trait = parent_controller.trait(control_name).inner_traits[0]
        fnames = None
        current_dir = os.path.join(os.getcwd(), os.pardir)
        if isinstance(elem_trait.trait_type, traits.Directory):

            # Create a dialog to select a directory
            fdialog = QtGui.QFileDialog(
                control_instance, "Open directories",
                current_dir)
            fdialog.setOptions(QtGui.QFileDialog.ShowDirsOnly |
                               QtGui.QFileDialog.DontUseNativeDialog)
            fdialog.setFileMode(QtGui.QFileDialog.Directory)
            fdialog.setModal(True)
            if fdialog.exec_():
                fnames = fdialog.selectedFiles()
        else:
            if elem_trait.output:
                fdialog = QtGui.QFileDialog(
                    control_instance, "Output files",
                    current_dir)
                fdialog.setOptions(QtGui.QFileDialog.DontUseNativeDialog)
                fdialog.setFileMode(QtGui.QFileDialog.AnyFile)
                fdialog.setModal(True)
                if fdialog.exec_():
                    fnames = fdialog.selectedFiles()
            else:
                fdialog = QtGui.QFileDialog(
                    control_instance, "Open files",
                    current_dir)
                fdialog.setOptions(QtGui.QFileDialog.DontUseNativeDialog)
                fdialog.setFileMode(QtGui.QFileDialog.ExistingFiles)
                fdialog.setModal(True)
                if fdialog.exec_():
                    fnames = fdialog.selectedFiles()

        # Set the selected files to the path sub control
        if fnames is not None:
            old_value = getattr(parent_controller, control_name)
            new_value = old_value + fnames
            setattr(parent_controller, control_name, new_value)


    @staticmethod
    def clear_all(controller_widget, control_name, control_instance, minlen):
        controller = control_instance.controller
        parent_controller = controller_widget.controller
        value = parent_controller.trait(control_name).default
        setattr(parent_controller, control_name, value)


class ListValuesEditor(QtGui.QDialog):

    def __init__(self, parent, controller_widget, control_name):

        super(ListValuesEditor, self).__init__(parent)

        self.controller_widget = controller_widget
        self.control_name = control_name
        self.format = 'JSON'
        self.separator = ','
        self.modified = False

        layout = QtGui.QVBoxLayout()
        self.setLayout(layout)
        textedit = QtGui.QTextEdit()
        layout.addWidget(textedit)
        hlayout2 = QtGui.QHBoxLayout()
        layout.addLayout(hlayout2)

        hlayout2.addWidget(QtGui.QLabel('Format:'))
        format_c = QtGui.QComboBox()
        hlayout2.addWidget(format_c)
        hlayout2.addWidget(QtGui.QLabel('Separator:'))
        sep_c = QtGui.QComboBox()
        hlayout2.addWidget(sep_c)

        format_c.addItem('JSON')
        format_c.addItem('CSV')

        sep_c.addItem(',')
        sep_c.addItem(';')
        sep_c.addItem(' ')

        hlayout2.addStretch(1)
        ok = QtGui.QPushButton('OK')
        cancel = QtGui.QPushButton('Cancel')
        hlayout2.addWidget(ok)
        hlayout2.addWidget(cancel)

        ok.pressed.connect(self.accept)
        cancel.pressed.connect(self.reject)

        parent_controller = controller_widget.controller
        value = getattr(parent_controller, control_name)

        text = json.dumps(value)
        textedit.setText(text)

        self.textedit = textedit
        self.format_c = format_c
        self.separator_c = sep_c
        self.internal_change = False

        textedit.textChanged.connect(self.set_modified)
        format_c.currentIndexChanged.connect(self.format_changed)
        sep_c.currentIndexChanged.connect(self.separator_changed)

    def set_modified(self):
        if self.internal_change:
            self.internal_change = False
            return
        self.modified = True
        self.textedit.textChanged.disconnect(self.set_modified)

    def format_changed(self, index):
        self.format = self.format_c.itemText(index)
        self.internal_change = True
        if self.modified:
            return
        self.textedit.setText(self.text_repr())

    def text_repr(self):
        parent_controller = self.controller_widget.controller
        value = getattr(parent_controller, self.control_name)
        if self.format == 'JSON':
            text = json.dumps(value)
        elif self.format == 'CSV':
            text_io = StringIO()
            writer = csv.writer(
                text_io, delimiter=str(self.separator),
                quotechar='"', quoting=csv.QUOTE_MINIMAL,
                skipinitialspace=True)
            writer.writerow(value)
            text = text_io.getvalue()
        else:
            raise ValueError('unsupported format %s' % self.format)
        return text

    def separator_changed(self, index):
        self.separator = self.separator_c.itemText(index)
        self.internal_change = True
        if self.modified:
            return
        self.textedit.setText(self.text_repr())



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''Utility functions to make a weak proxy which also keeps an access to its original object reference. :func:`weakref.proxy` doesn't allow this, but functions that check types (C+/Python bindings for instance) cannot work with proxies.

We build such a proxy by setting a :func:`weakref.ref` object in the proxy (actually in the object itself).
'''

from __future__ import absolute_import
import weakref

def get_ref(obj):
    ''' Get a regular reference to an object, whether it is already a regular
    reference, a weak reference, or a weak proxy which holds an access to the
    original reference (built using :func:`weak_proxy`).
    In case of a weak proxy not built using :func:`weak_proxy`, we try to get
    the ``self`` from a bound method of the object, namely
    ``obj.__init__.__self__``, if it exists.
    '''
    if isinstance(obj, weakref.ReferenceType):
        return obj()
    elif isinstance(obj, weakref.ProxyTypes):
        if hasattr(obj, '_weakref'):
            return obj._weakref()
        elif hasattr(obj, '__init__'):
            # try to get the 'self' of a bound method
            return obj.__init__.__self__
    return obj


def weak_proxy(obj, callback=None):
    ''' Build a weak proxy (:class:`weakref.ProxyType`) from an object, if it
    is not already one, and keep a reference to the original object (via a
    :class:`weakref.ReferenceType`) in it.

    *callback* is passed to :func:`weakref.proxy`.
    '''
    if isinstance(obj, weakref.ProxyTypes):
        return obj
    real_obj = get_ref(obj)
    if callback:
        wr = weakref.proxy(real_obj, callback)
    else:
        wr = weakref.proxy(real_obj)
    wr._weakref = weakref.ref(real_obj)
    return wr


class proxy_method(object):
    ''' Indirect proxy for a bound method

    It replaces a bound method, ie ``a.method`` with a proxy callable which
    does not take a reference on ``a``.

    Especially useful for callbacks.
    If we want to set a notifier with a callback on a proxy object (without
    adding a new reference on it), we can use proxy_method::

        a = anatomist.Anatomist()
        a.onCursorNotifier.onAddFirstListener.add(
            partial(proxy_method(a, 'enableListening'),
            "LinkedCursor", a.onCursorNotifier)))
        del a

    Without this mechanism, using::

        a.onCursorNotifier.onAddFirstListener.add(
            partial(a.enableListening, "LinkedCursor", a.onCursorNotifier)))

    would increment the reference count on a, because ``a.enableListening``,
    as a *bound method*, contains a reference to a, and will prevent the
    deletion of ``a`` (here the Anatomist application)
    '''
    def __init__(self, obj, method=None):
        '''
        The constructor takes as parameters, either the object and its method
        name (as a string), or the bound method itself.
        '''
        if method is None:
            method = obj.__name__
            obj = obj.__self__
        self.proxy = weak_proxy(obj)
        self.method = method

    def __call__(self, *args, **kwargs):
        return getattr(self.proxy, self.method)(*args, **kwargs)



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
* author: Yann Cointepas
* organization: NeuroSpin
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''

from __future__ import absolute_import
__docformat__ = "restructuredtext en"


import os
import sys
import platform
import traceback
from os.path import dirname

os.environ['ETS_TOOLKIT'] = 'qt4'
try:
    from enthought.traits.api import ReadOnly, Directory, ListStr, Instance
except ImportError:
    from traits.api import ReadOnly, Directory, ListStr, Instance

from soma.singleton import Singleton
from soma.controller import Controller

#-------------------------------------------------------------------------


class Application(Singleton, Controller):

    '''Any program using soma should create an Application instance to manage
    its configuration and to store any kind of value that have to be global to
    the program.'''
    name = ReadOnly(desc='Name of the application')
    version = ReadOnly()

    plugin_modules = ListStr(
        desc='List of Python module to load after application configuration')

    def __singleton_init__(self, name=None, version=None, *args, **kwargs):
        '''Replaces __init__ in Singleton.'''
        super(Application, self).__singleton_init__(*args, **kwargs)
        # Warning : Traits bug
        # Using the trait Directory() might instanciate a QApplication (seems to depend on the
        # traits release). If it is declared in the class, the QApplication is instanciated at
        # module importation which prevent to customize QApplication.
        self.add_trait('install_directory', Directory(
                       desc='Base directory where the application is installed'))
        self.add_trait('user_directory', Directory(
                       desc='Base directory where user specific information can be find'))
        self.add_trait('application_directory', Directory(
                       desc='Base directory where application specifc information can be find'))
        self.add_trait('site_directory',  Directory(
                       desc='Base directory where site specifc information can be find'))
        self._controller_factories = None

        if name is None:
            name = os.path.basename(sys.argv[0])
        self.name = name
        self.version = version
        self.loaded_plugin_modules = {}

    def initialize(self):
        '''This method must be called once to setup the application.'''
        self.install_directory = dirname(dirname(dirname(__file__)))
        homedir = os.getenv('HOME')
        if not homedir:
            homedir = ''
            if platform.system() == 'Windows':
                homedir = os.getenv('USERPROFILE')
                if not homedir:
                    homedir = os.getenv('HOMEPATH')
                    if not homedir:
                        homedir = '\\'
                    drive = os.getenv('HOMEDRIVE')
                    if not drive:
                        drive = os.getenv('SystemDrive')
                        if not drive:
                            drive = os.getenv('SystemRoot')
                            if not drive:
                                drive = os.getenv('windir')
                            if drive and len(drive) >= 2:
                                drive = drive[:2]
                            else:
                                drive = ''
                    homedir = drive + homedir
        if homedir and os.path.exists(homedir):
            self.user_directory = homedir

        # Load early plugin modules
        for plugin_module in self.plugin_modules:
            module = self.load_plugin_module(plugin_module)
            if module is not None:
                self.loaded_plugin_modules[plugin_module] = module
                init = getattr(
                    module, 'call_before_application_initialization', None)
                if init is not None:
                    init(self)

        appdir = os.path.normpath(
            os.path.dirname(os.path.dirname(sys.argv[0])))
        if os.path.exists(appdir):
            self.application_directory = appdir

        sitedir = os.path.join('/etc', self.name)
        if os.path.exists(sitedir):
            self.site_directory = sitedir

        # Load plugin modules
        for plugin_module in self.plugin_modules:
            module = self.loaded_plugin_modules.get(plugin_module)
            if module is None:
                module = self.load_plugin_module(plugin_module)
            if module is not None:
                self.loaded_plugin_modules[plugin_module] = module
                init = getattr(
                    module, 'call_after_application_initialization', None)
                if init is not None:
                    init(self)

    @staticmethod
    def load_plugin_module(plugin_module):
        '''This method loads a plugin module. It imports the module without raising
        an exception if it fails.'''
        try:
            __import__(plugin_module, level=0)
            return sys.modules[plugin_module]
        except Exception:
            # Python 2.6 hack : print_last may fail here (maybe due to threads management)
            # traceback.print_last()
            exceptionType, exceptionValue, exceptionTraceback = sys.exc_info()
            traceback.print_exception(
                exceptionType, exceptionValue, exceptionTraceback)
        return None



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
This module contains the L{XHTML} class that contains an XHTML tree that
can be saved in minf files.

* author: Yann Cointepas
* organization: `NeuroSpin <http://www.neurospin.org>`_
* license: `CeCILL B <http://www.cecill.info/licences/Licence_CeCILL-B_V1-en.html>`_
'''
from __future__ import absolute_import
__docformat__ = "restructuredtext en"

import types
try:
    from cStringIO import StringIO
except ImportError:
    # python3
    from io import StringIO

import six
import sys

from soma.translation import translate as _
from soma.minf.api import readMinf
from soma.minf.xml_tags import minfTag, expanderAttribute, xhtmlTag
from soma.html import lesserHtmlEscape
from xml.sax.saxutils import quoteattr as xml_quoteattr

#------------------------------------------------------------------------------


class XHTML(object):

    '''
    Instances of L{XHTML} contains the structure of an XHTML tree and can be used
    to produce an XML string or an HTML string.
    '''

    def __init__(self, tag, attributes=None, content=None):
        '''
        Construct an XHTML tree composed of a tag name, a dictionary containing
        attributes and a content composed of a series of strings and/or XHTML
        values.
        '''
        if attributes is None:
            attributes = {}
        if content is None:
            content = []
        self.tag = tag
        self.attributes = attributes
        self.content = content

    def __getinitkwargs__(self):
        d = {}
        if self.attributes:
            d['attributes'] = self.attributes
        if self.content:
            d['content'] = self.content
        return (self.tag, ), d

    def _contentXML(content):
        return ''.join([XHTML._itemXML(i) for i in content])
    _contentXML = staticmethod(_contentXML)

    def _itemXML(item):
        if isinstance(item, XHTML):
            if item.tag is None:
                # When tag is None, no tag info is produced
                # This is useful for creating XHTML converters
                # that can remove tags.
                return item._contentXML(item.content)
            else:
                result = '<' + six.text_type(item.tag)
                if item.attributes:
                    result += ' ' + ' '.join(
                        [six.text_type(a) + '="' + six.text_type(v) + '"'
                         for a, v in six.iteritems(item.attributes)])
                if item.content:
                    result += '>' + item._contentXML( item.content ) + '</' + \
                              str(item.tag) + '>'
                else:
                    result += '/>'
                return result
        else:
            return lesserHtmlEscape(item)
    _itemXML = staticmethod(_itemXML)

    def __minfxml__(self, xmlWriter, attributes, level):
        backupAttributes = self.attributes
        self.attributes = self.attributes.copy()
        self.attributes.update(attributes)
        xmlWriter._encodeAndWriteLine(self._itemXML(self), level)
        self.attributes = backupAttributes

    def buildFromHTML(html):
        '''
        Return an XHTML instance build from an HTML string.
        @param html: piece of an HTML file
        @type  html: unicode
        '''
        # html must be an unicode string. if not, we can fail to decode it because we don't know what encoding has been use to encode it.
        # try:
            # if not isinstance( html, unicode ):
                # html = unicode( html) #, 'iso-8859-1' )
        # except: # decoding of the string can fail because it isn't encoded with the default charset
                # pass
        io = StringIO()
        # when unicode string is written in a stream, default encoding is used
        # to encode it :
        html = six.ensure_str(html, 'utf-8')
        io.write(
            '<?xml version="1.0" encoding="utf-8" ?>\n<' + minfTag + ' ' +
            expanderAttribute + '="minf_2.0">\n<' + xhtmlTag + '>' +
            html + '</' + xhtmlTag + '>\n</' + minfTag + '>')
        io.seek(0)
        return readMinf(io)[0]
    buildFromHTML = staticmethod(buildFromHTML)

    def xml(item):
        '''
        Build an XML unicode string based on either an XML string (in that case it
        is returned as is) or an XHTML instance.
        @param item: value to convert in XML.
        @type  item: XHTML instance or unicode containing XML
        '''
        if isinstance(item, six.string_types):
            return item
        elif isinstance(item, XHTML):
            return item._itemXML(item)
        else:
            raise RuntimeError(
                _('Cannot use XHTML converter for %s') % (str(item),))
    xml = staticmethod(xml)

    def html(item):
        '''
        Build an HTML unicode string based on either an HTML string (in that case it
        is returned as is) or an XHTML instance.
        @param item: value to convert in HTML.
        @type  item: XHTML instance or unicode containing HTML
        '''
        if isinstance(item, six.string_types):
            return item
        elif isinstance(item, XHTML):
            return item._contentXML(item.content)
        else:
            raise RuntimeError(
                _('Cannot use XHTML converter for %s') % (six.text_type(item),))
    html = staticmethod(html)



# -*- coding: utf-8 -*-

#  This software and supporting documentation are distributed by
#      Institut Federatif de Recherche 49
#      CEA/NeuroSpin, Batiment 145,
#      91191 Gif-sur-Yvette cedex
#      France
#
# This software is governed by the CeCILL-B license under
# French law and abiding by the rules of distribution of free software.
# You can  use, modify and/or redistribute the software under the
# terms of the CeCILL-B license as circulated by CEA, CNRS
# and INRIA at the following URL "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL-B license and that you accept its terms.

'''
Env variables parsing tools
'''


def parse_env_lines(text, asdict=False):
    ''' Separate text (generally the output of the ``env`` command) into multi-
    line elements, avoiding separations inside () or {} blocks.

    If ``asdict`` is ``True``, then the result is returned as a dict
    ``{variable: value}`` (see :func:`env_to_dict`)
    '''
    def push(obj, l, depth, tags, start_tag=None):
        while depth:
            l = l[-1]
            tags = tags[-1]
            depth -= 1

        if start_tag:
            tags.append([start_tag, []])
        if isinstance(obj, str):
            if len(l) == 0 or isinstance(l[-1], list):
                l.append(obj)
            else:
                l[-1] += obj
        else:
            l.append(obj)

    def current_tag(tags, depth):
        while depth:
            tags = tags[-1]
            depth -= 1
        return tags[0]

    def parse_parentheses(s):
        rev_char = {'(': ')', '{': '}', #'[': ']',
                    '"': '"', "'": "'"}
        groups = []
        tags = [None, []]
        depth = 0
        escape = None

        try:
            for char in s:
                if s == '\\':
                    escape = not escape
                    if escape:
                       push(char, groups, depth, tags)

                if char == rev_char.get(current_tag(tags, depth)):
                    # close tag (counterpart of the tag)
                    push(char, groups, depth, tags)
                    depth -= 1

                elif char in rev_char:
                    # open/start tag
                    push([char], groups, depth, tags, char)
                    depth += 1

                else:
                    push(char, groups, depth, tags)

        except IndexError:
            raise ValueError('Parentheses mismatch', depth, groups)
        if depth > 0:
            raise ValueError('Parentheses mismatch 2', depth, groups)
        else:
            return groups

    def rebuild_lines(parsed, breaks=True):
        lines = []
        for item in parsed:
            if isinstance(item, str):
                if breaks:
                    newlines = item.split('\n')
                else:
                    newlines = [item]
            else:
                newlines = rebuild_lines(item, breaks=False)
            if lines:
                lines[-1] += newlines[0]
                lines += newlines[1:]
            else:
                lines = newlines
        return lines

    lines = rebuild_lines(parse_parentheses(text))
    if asdict:
        lines = env_to_dict(lines)

    return lines


def env_to_dict(lines):
    ''' Separate each text line in the lines list into variable: value.
    Each line is expected to be in the shape ``VAR=value``, as expected from
    the output of the ``env`` command, parsed by the :func:`parse_env_lines`
    function. Note that :func:`parse_env_lines` with the option ``asdict=True``
    will already call this function.
    '''
    denv = {}
    for line in lines:
        var, value = line.split('=', 1)
        denv[var.strip()] = value.strip()

    return denv
